import os
from busco.BuscoConfig import BuscoConfigAuto
from busco.BuscoPlacer import BuscoPlacer
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from busco.BuscoRunner import BuscoRunner
import numpy as np

logger = BuscoLogger.get_logger(__name__)


class NoGenesError(SystemExit):

    def __init__(self):
        super().__init__("No genes were recognized by BUSCO. Please check the content of your input file.")


class AutoSelectLineage:
    """
    Class for selecting the best lineage dataset for the input data.
    Auto Selector works by running BUSCO using all available datasets and identifying the dataset that returns the
    highest BUSCO score.
    """

    runners = []

    @log("***** Starting Auto Select Lineage *****\n\t"
         "This process runs BUSCO on the generic lineage datasets for the domains archaea, bacteria and eukaryota. "
         "Once the optimal domain is selected, BUSCO automatically attempts to find the most appropriate BUSCO dataset "
         "to use based on phylogenetic placement.\n\t"
         "--auto-lineage-euk and --auto-lineage-prok are also available if you know your input assembly is, or is not, an"
         " eukaryote. See the user guide for more information.\n\tA reminder: Busco evaluations are valid when an "
         "appropriate dataset is used, i.e., the dataset belongs to the lineage of the species to test. "
         "Because of overlapping markers/spurious matches among domains, busco matches in another domain do not "
         "necessarily mean that your genome/proteome contains sequences from this domain. "
         "However, a high busco score in multiple domains might help you identify possible contaminations.", logger)
    def __init__(self, config):
        self.config = config
        if self.config.getboolean("busco_run", "auto-lineage-prok"):
            self.all_lineages = ["archaea", "bacteria"]
        elif self.config.getboolean("busco_run", "auto-lineage-euk"):
            self.all_lineages = ["eukaryota"]
        else:
            self.all_lineages = ["archaea", "bacteria", "eukaryota"]
        self.dataset_version = self.config.get("busco_run", "datasets_version")
        self.callback = self.record_results
        self.s_buscos = []
        self.d_buscos = []
        self.f_buscos = []
        self.s_percents = []
        self.d_percents = []
        self.f_percents = []
        self.best_match_lineage_dataset = None
        self.current_lineage = None
        self.virus_pipeline = False

    def record_results(self, s_buscos, d_buscos, f_buscos, s_percent, d_percent, f_percent):
        """
        Record results of BUSCO run.

        :param float s_buscos: Number of Single copy BUSCOs present in data
        :param float d_buscos: Number of Double copy BUSCOs present in data
        :param float f_buscos: Number of Fragmented BUSCOs present in data
        :return: None
        """
        self.s_buscos.append(s_buscos)
        self.d_buscos.append(d_buscos)
        self.f_buscos.append(f_buscos)
        self.s_percents.append(s_percent)
        self.d_percents.append(d_percent)
        self.f_percents.append(f_percent)
        return

    @log("Running auto selector", logger, debug=True)
    def run_auto_selector(self):
        """
        Run BUSCO on each lineage listed in all_lineages.
        :return:
        """
        root_runners = self.run_lineages_list(self.all_lineages)
        self.get_best_match_lineage(root_runners)
        if self.virus_check():
            self.virus_pipeline = True
            self.run_virus_datasets()
            self.get_best_match_lineage(type(self).runners)

        if (len(self.selected_runner.analysis.hmmer_runner.single_copy_buscos) ==
                len(self.selected_runner.analysis.hmmer_runner.multi_copy_buscos) ==
                len(self.selected_runner.analysis.hmmer_runner.fragmented_buscos) == 0):
            raise NoGenesError

        logger.info("{} selected\n".format(os.path.basename(self.best_match_lineage_dataset)))
        self.config.set("busco_run", "domain_run_name", os.path.basename(self.best_match_lineage_dataset))
        BuscoRunner.final_results.append(self.selected_runner.analysis.hmmer_runner.hmmer_results_lines)
        BuscoRunner.results_datasets.append(os.path.basename(self.best_match_lineage_dataset))
        return

    def virus_check(self):
        return (self.selected_runner.analysis.hmmer_runner.s_percent < 3.0) & \
               (os.stat(self.selected_runner.analysis.input_file).st_size < 500000)

    @log("Running virus detection pipeline", logger)
    def run_virus_datasets(self):
        lineages_to_check = []
        virus_datasets = self.selected_runner.config.downloader.get("virus_datasets.txt", "information")
        with open(virus_datasets, "r") as vir_sets:
            for line in vir_sets:
                lineages_to_check.append(line.strip().split("_odb")[0])
        self.run_lineages_list(lineages_to_check)
        return


    def run_lineages_list(self, lineages_list):
        root_runners = []
        for l in lineages_list:
            self.current_lineage = "{}_{}".format(l, self.dataset_version)
            autoconfig = BuscoConfigAuto(self.config, self.current_lineage)
            busco_run = BuscoRunner(autoconfig)
            busco_run.run_analysis(callback=self.callback)
            root_runners.append(busco_run)
            type(self).runners.append(busco_run)  # Save all root runs so they can be recalled if chosen
        return root_runners

    def get_max_ind(self, arr):
        """
        Return maximum ind(s) of array. If max value appears twice, two indices are returned.
        :param arr:
        :return:
        """
        inds = np.arange(len(arr))
        max_mask = arr == np.amax(arr)
        max_ind = inds[max_mask]
        return max_ind

    def evaluate(self, runners):
        """
        Evaluate output scores from all BUSCO runs. Lineage with the highest number of complete (single + multiple)
        copy BUSCOs is assigned as the best_match_lineage.
        In case of a tie, the number of fragmented BUSCOs are used as a tiebreak.
        I this is still a tie and the number of matched BUSCOs is zero, then an error is raised.
        If there is a further nonzero tie, the tiebreak is the highest percentage of single copy BUSCOs.
        If still a tie, use the first match.
        :return
        """
        self.collate_results(runners)

        max_ind = self.get_max_ind(np.array(self.s_buscos) + np.array(self.d_buscos))
        if len(max_ind) > 1:
            max_ind2 = self.get_max_ind(np.array(self.f_buscos)[max_ind])
            max_ind = max_ind[max_ind2]
            if len(max_ind) > 1:
                if ((self.s_buscos[max_ind[0]] == 0.0)
                        and (self.d_buscos[max_ind[0]] == 0.0)
                        and (self.f_buscos[max_ind[0]] == 0.0)):
                    return int(0)
                else:
                    max_ind3 = self.get_max_ind(np.array(self.s_percents)[max_ind])
                    max_ind = max_ind[max_ind3]
                    if len(max_ind) > 1:
                        logger.warning("Two lineage runs scored exactly the same. Proceeding with the first.")  # I don't expect this error message will ever be used.
                        max_ind = max_ind[0]

        return int(max_ind)

    def collate_results(self, runners):
        self.s_buscos = [runner.analysis.hmmer_runner.single_copy for runner in runners]
        self.d_buscos = [runner.analysis.hmmer_runner.multi_copy for runner in runners]
        self.f_buscos = [runner.analysis.hmmer_runner.only_fragments for runner in runners]
        self.s_percents = [runner.analysis.hmmer_runner.s_percent for runner in runners]
        return

    def get_best_match_lineage(self, runners):
        max_ind = self.evaluate(runners)
        self.selected_runner = runners[int(max_ind)]
        self.best_match_lineage_dataset = self.selected_runner.config.get("busco_run", "lineage_dataset")
        runners.pop(int(max_ind))
        self.cleanup_disused_runs(runners)
        return

    def cleanup_disused_runs(self, disused_runners):
        for runner in disused_runners:
            if not runner.cleaned_up:
                runner.cleanup()


    def get_lineage_dataset(self):  # todo: rethink structure after BuscoPlacer is finalized and protein mode with mollicutes is fixed.
        """
        Run the output of the auto selection through BuscoPlacer to obtain a more precise lineage dataset.
        :return str lineage_dataset: Local path to the optimal lineage dataset.
        """
        if self.selected_runner.domain == "eukaryota":
            self.run_busco_placer()
        elif (self.selected_runner.mode in ["proteins", "prok_tran"] and
              os.path.basename(self.selected_runner.config.get("busco_run", "lineage_dataset")).startswith("bacteria")):
            logger.info(
                "Certain mollicute clades use a different genetic code to the rest of bacteria. They are not part "
                "of the BUSCO placement tree and need to be tested separately. For more information, see the user "
                "guide.")
            self.check_mollicutes()
            if os.path.basename(self.selected_runner.config.get("busco_run", "lineage_dataset")).startswith("bacteria"):
                logger.info("Bacteria domain is a better match than the mollicutes subclade. Continuing to tree placement.")
                self.run_busco_placer()
            else:
                logger.info("Mollicutes dataset is a better match for your data. Testing subclades...")
                self._run_3_datasets(self.selected_runner)
                BuscoRunner.final_results.append(self.selected_runner.analysis.hmmer_runner.hmmer_results_lines)
                BuscoRunner.results_datasets.append(os.path.basename(self.best_match_lineage_dataset))
        elif ("geno" in self.selected_runner.mode
              and self.selected_runner.analysis.prodigal_runner.current_gc == "4"
              and os.path.basename(
                    self.selected_runner.config.get("busco_run", "lineage_dataset")).startswith("bacteria")):
            logger.info("The results from the Prodigal gene predictor indicate that your data belongs to the "
                        "mollicutes clade. Testing subclades...")
            self._run_3_datasets()
            BuscoRunner.final_results.append(self.selected_runner.analysis.hmmer_runner.hmmer_results_lines)
            BuscoRunner.results_datasets.append(os.path.basename(self.best_match_lineage_dataset))
        elif self.selected_runner.domain == "viruses":
            pass
        else:
            self.run_busco_placer()
        return

    def check_mollicutes(self):
        runners = self.run_lineages_list(["mollicutes"])
        runners.append(self.selected_runner)
        self.get_best_match_lineage(runners)
        return

    def run_busco_placer(self):  # todo: revisit structure of this method after cleaning BuscoPlacer
        if "genome" in self.selected_runner.mode:
            if self.selected_runner.domain == "prokaryota":
                protein_seqs = self.selected_runner.analysis.prodigal_runner.output_faa
            elif self.selected_runner.domain == "eukaryota":
                if self.config.getboolean("busco_run", "use_augustus"):
                    protein_seqs_dir = self.selected_runner.analysis.augustus_runner.extracted_prot_dir
                    protein_seqs = [os.path.join(protein_seqs_dir, f) for f in os.listdir(protein_seqs_dir)
                                    if f.split(".")[-2] == "faa"]
                else:
                    protein_seqs = self.selected_runner.analysis.metaeuk_runner.combined_pred_protein_seqs
        elif "tran" in self.selected_runner.mode:
            if self.selected_runner.mode == "euk_tran":
                protein_seqs = self.selected_runner.analysis.metaeuk_runner.combined_pred_protein_seqs
            elif self.selected_runner.mode == "prok_tran":
                protein_seqs = self.selected_runner.analysis.single_copy_proteins_file

        else:
            protein_seqs = self.selected_runner.config.get("busco_run", "in")
        out_path = self.config.get("busco_run", "main_out")
        run_folder = os.path.join(out_path, "auto_lineage", self.selected_runner.config.get("busco_run", "lineage_results_dir"))
        bp = BuscoPlacer(self.selected_runner.config, run_folder, protein_seqs, self.selected_runner.analysis.hmmer_runner.single_copy_buscos)
        dataset_details, placement_file_versions = bp.define_dataset()
        self.config.placement_files = placement_file_versions  # Necessary to pass these filenames to the final run to be recorded.
        lineage, supporting_markers, placed_markers = dataset_details
        lineage = "{}_{}".format(lineage, self.config.get("busco_run", "datasets_version"))  # todo: this should probably be done in buscoplacer
        self.best_match_lineage_dataset = os.path.join(self.config.get("busco_run", "download_path"),
                                                       "lineages",
                                                       os.path.basename(lineage))
        return

    def _run_3_datasets(self, mollicutes_runner=None):
        if mollicutes_runner:
            datasets = ["mycoplasmatales", "entomoplasmatales"]
            dataset_runners = [mollicutes_runner]
        else:
            datasets = ["mollicutes", "mycoplasmatales", "entomoplasmatales"]
            dataset_runners = []
        dataset_runners += self.run_lineages_list(datasets)
        self.get_best_match_lineage(dataset_runners)
        return


from busco.busco_tools.base import BaseRunner, NoRerunFile, NoGenesError
import os
from collections import defaultdict
from busco.BuscoLogger import BuscoLogger
from Bio import SeqIO
import shutil
from configparser import NoOptionError
import subprocess
import gzip
import pandas as pd
import numpy as np
import re

logger = BuscoLogger.get_logger(__name__)


class MetaeukParsingError(Exception):
    def __init__(self):
        pass


class MetaeukRunner(BaseRunner):

    name = "metaeuk"
    cmd = "metaeuk"

    ACCEPTED_PARAMETERS = [
        "comp-bias-corr",
        "add-self-matches",
        "seed-sub-mat",
        "s",
        "k",
        "k-score",
        "alph-size",
        "max-seqs",
        "split",
        "split-mode",
        "split-memory-limit",
        "diag-score",
        "exact-kmer-matching",
        "mask",
        "mask-lower-case",
        "min-ungapped-score",
        "spaced-kmer-mode",
        "spaced-kmer-pattern",
        "local-tmp",
        "disk-space-limit",
        "a",
        "alignment-mode",
        "wrapped-scoring",
        "e",
        "min-seq-id",
        "min-aln-len",
        "seq-id-mode",
        "alt-ali",
        "c",
        "cov-mode",
        "realign",
        "max-rejected",
        "max-accept",
        "score-bias",
        "gap-open",
        "gap-extend",
        "zdrop",
        "pca",
        "pcb",
        "mask-profile",
        "e-profile",
        "wg",
        "filter-msa",
        "max-seq-id",
        "qid",
        "qsc",
        "cov",
        "diff",
        "num-iterations",
        "slice-search",
        "rescore-mode",
        "allow-deletion",
        "min-length",
        "max-length",
        "max-gaps",
        "contig-start-mode",
        "contig-end-mode",
        "orf-start-mode",
        "forward-frames",
        "reverse-frames",
        "translation-table",
        "translate",
        "use-all-table-starts",
        "id-offset",
        "add-orf-stop",
        "search-type",
        "start-sens",
        "sens-steps",
        "metaeuk-eval",
        "metaeuk-tcov",
        "min-intron",
        "min-exon-aa",
        "max-overlap",
        "set-gap-open",
        "set-gap-extend",
        "overlap",
        "protein",
        "target-key",
        "reverse-fragments",
        "sub-mat",
        "db-load-mode",
        "force-reuse",
        "remove-tmp-files",
        "filter-hits",
        "sort-results",
        "omit-consensus",
        "create-lookup",
        "chain-alignments",
        "merge-query",
        "strand",
        "compressed",
        "v",
        "max-intron",
        "max-seq-len",
    ]

    PRESET_PARAMETERS = [
        "--max-intron",
        "--max-seq-len",
        "--min-exon-aa",
        "--max-overlap",
        "--min-intron",
        "--overlap",
    ]

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(self.run_folder, "metaeuk_output")
        self._initial_results_folder = os.path.join(
            self._output_folder, "initial_results"
        )
        self._rerun_results_folder = os.path.join(self._output_folder, "rerun_results")
        self._tmp_folder = os.path.join(self._output_folder, "tmp")

        self.ancestral_file = os.path.join(self.lineage_dataset, "ancestral")
        self.ancestral_variants_file = os.path.join(
            self.lineage_dataset, "ancestral_variants"
        )

        self.max_intron = self.config.get("busco_run", "max_intron")
        self.max_seq_len = self.config.get("busco_run", "max_seq_len")
        self.overlap = 1
        self.s_set = False

        self.extra_params = None
        self.param_keys = []
        self.param_values = []
        self.create_dirs(
            [
                self._output_folder,
                self._initial_results_folder,
                self._rerun_results_folder,
            ]
        )
        self.gene_details = None

        self.headers_file = None
        self.codon_file = None
        self.pred_protein_seqs = None
        self.pred_protein_seqs_modified = None
        self.incomplete_buscos = None
        self.sequences_aa = {}
        self.init_checkpoint_file()

        self.pred_protein_files = []
        self.pred_protein_mod_files = []
        self.headers_files = []
        self.combined_pred_protein_seqs = os.path.join(
            self._output_folder, "combined_pred_proteins.fas"
        )

    def configure_runner(self, incomplete_buscos=None):
        self.run_number += 1
        self.incomplete_buscos = incomplete_buscos

        if self.run_number > 1:
            self._output_basename = os.path.join(
                self._rerun_results_folder, os.path.basename(self.input_file)
            )
            self.refseq_db_rerun = os.path.join(
                self._output_folder, "refseq_db_rerun.faa"
            )
            self._extract_incomplete_buscos_ancestral()
            self.refseq_db = self.refseq_db_rerun
            self.min_exon_aa = 5
            self.max_overlap = 5
            self.min_intron = 1
        else:
            self._output_basename = os.path.join(
                self._initial_results_folder, os.path.basename(self.input_file)
            )
            gzip_refseq = os.path.join(self.lineage_dataset, "refseq_db.faa.gz")
            self.refseq_db = self.decompress_refseq_file(gzip_refseq)
            self.min_exon_aa = 15
            self.max_overlap = 15
            self.min_intron = 5

        try:
            if self.run_number == 1:
                self.extra_params = self.config.get(
                    "busco_run", "metaeuk_parameters"
                ).replace(",", " ")
            else:
                self.extra_params = self.config.get(
                    "busco_run", "metaeuk_rerun_parameters"
                ).replace(",", " ")
        except NoOptionError:
            self.extra_params = ""

        self.headers_file = "{}.headersMap.tsv".format(self._output_basename)
        self.headers_files.append(self.headers_file)
        self.codon_file = "{}.codon.fas".format(self._output_basename)
        self.pred_protein_seqs = "{}.fas".format(self._output_basename)
        self.pred_protein_files.append(self.pred_protein_seqs)
        self.pred_protein_seqs_modified = self.pred_protein_seqs.replace(
            ".fas", ".modified.fas"
        )
        self.pred_protein_mod_files.append(self.pred_protein_seqs_modified)

    def decompress_refseq_file(self, gzip_file):
        unzipped_filename = gzip_file.split(".gz")[0]
        if not os.path.exists(unzipped_filename):
            with gzip.open(gzip_file, "rb") as compressed_file:
                with open(unzipped_filename, "wb") as decompressed_file:
                    for line in compressed_file:
                        decompressed_file.write(line)
        if os.path.exists(gzip_file):
            try:
                os.remove(gzip_file)
            except OSError:
                logger.warning(
                    "Unable to remove compressed refseq file in dataset download"
                )
                pass
        return unzipped_filename

    def combine_run_results(self):
        with open(self.combined_pred_protein_seqs, "w") as combined_output:
            for run_result in self.pred_protein_mod_files:
                with open(run_result, "r") as f:
                    shutil.copyfileobj(f, combined_output)

        return

    def records_to_df(self, records):
        results = pd.DataFrame.from_records(
            records,
            columns=[
                "Busco id",
                "Sequence",
                "Start",
                "Stop",
                "Strand",
                "Score",
                "Run found",
            ],
            index=np.arange(len(records)),
        )

        return results

    @staticmethod
    def detect_overlap(match1_details, match2_details):
        return (
            match1_details["Strand"] == match2_details["Strand"]
        )  # check overlaps are on the same strand

    def test_for_overlaps(self, record_df):
        results_grouped = record_df.groupby("Sequence")
        overlaps = []
        seq_ids = results_grouped.groups.keys()
        for seq in seq_ids:
            g1 = results_grouped.get_group(seq)
            g1_sorted = g1.sort_values(
                "Start"
            )  # sort to facilitate a single-pass coordinate check
            for idx1, row1 in g1_sorted.iterrows():
                start_val = g1_sorted.loc[idx1]["Start"]
                stop_val = g1_sorted.loc[idx1]["Stop"]
                matches = g1_sorted[g1_sorted["Start"] > start_val].loc[
                    g1_sorted["Start"] < stop_val
                ]  # find entries with a start coordinate between the current exon start and end
                for idx2 in matches.index.values:
                    if self.detect_overlap(g1_sorted.loc[idx1], g1_sorted.loc[idx2]):
                        overlaps.append((idx1, idx2))

        return overlaps

    def find_match(self, matches, id_items):
        matches_list = str(matches).split("\n")
        good_match = None
        good_matches = []
        for match in matches_list:
            if all([i in match for i in id_items]):
                good_match = match
                good_matches.append(good_match)
        if len(good_matches) > 1:
            good_ind = self.select_higher_bitscore_ind(good_matches)
            good_match = good_matches[good_ind]
        return good_match

    @staticmethod
    def select_higher_bitscore_ind(matches):
        bitscores = []
        pos_pattern = "|+|"
        neg_pattern = "|-|"
        for match in matches:
            strand_pattern = pos_pattern if pos_pattern in match else neg_pattern
            score = re.search(
                r"{}(.*?)\|".format(re.escape(strand_pattern)), match
            ).group(1)
            bitscores.append(int(score))
        max_ind = bitscores.index(max(bitscores))
        return max_ind

    def extract_exon_coords(self, match):
        parts = str(match).split("\t")
        header = parts[5]
        details = self.parse_header(header)
        return (
            details["all_taken_low_exon_coords"],
            details["all_taken_high_exon_coords"],
        )

    def check_tool_dependencies(self):
        pass

    def configure_job(self, *args):

        metaeuk_job = self.create_job()
        metaeuk_job.add_parameter("easy-predict")
        metaeuk_job.add_parameter("--threads")
        metaeuk_job.add_parameter(str(self.cpus))
        metaeuk_job.add_parameter(self.input_file)
        metaeuk_job.add_parameter(self.refseq_db)
        metaeuk_job.add_parameter(self._output_basename)
        metaeuk_job.add_parameter(self._tmp_folder)
        metaeuk_job.add_parameter("--max-intron")
        metaeuk_job.add_parameter(str(self.max_intron))
        metaeuk_job.add_parameter("--max-seq-len")
        metaeuk_job.add_parameter(str(self.max_seq_len))
        metaeuk_job.add_parameter("--min-exon-aa")
        metaeuk_job.add_parameter(str(self.min_exon_aa))
        metaeuk_job.add_parameter("--max-overlap")
        metaeuk_job.add_parameter(str(self.max_overlap))
        metaeuk_job.add_parameter("--min-intron")
        metaeuk_job.add_parameter(str(self.min_intron))
        metaeuk_job.add_parameter("--overlap")
        metaeuk_job.add_parameter(str(self.overlap))
        if self.run_number > 1 and self.s_set == False:
            metaeuk_job.add_parameter("-s")
            metaeuk_job.add_parameter("6")
        for k, key in enumerate(self.param_keys):
            dashes = "-" if len(key) == 1 else "--"
            metaeuk_job.add_parameter("{}{}".format(dashes, key))
            metaeuk_job.add_parameter("{}".format(str(self.param_values[k])))

        return metaeuk_job

    def generate_job_args(self):
        yield

    @property
    def output_folder(self):
        return self._output_folder

    def remove_tmp_files(self):
        shutil.rmtree(self._tmp_folder)

    def run(self):
        super().run()
        if self.extra_params:
            logger.info(
                "Additional parameters for Metaeuk are {}: ".format(self.extra_params)
            )
            self.param_keys, self.param_values = self.parse_parameters()

        # self.cwd = self._output_folder
        self.total = 1
        self.run_jobs()

    def _extract_incomplete_buscos_ancestral(self):

        logger.info(
            "Extracting missing and fragmented buscos from the file {}...".format(
                os.path.basename(self.refseq_db)
            )
        )

        matched_seqs = []
        busco_ids_retrieved = set()
        with open(self.refseq_db, "rU") as refseq_file:

            for record in SeqIO.parse(refseq_file, "fasta"):
                if any(record.id.startswith(b) for b in self.incomplete_buscos):
                    # Remove the ancestral variant identifier ("_1" etc) so it matches all other BUSCO IDs.
                    # The identifier is still present in the "name" and "description" Sequence Record attributes.
                    record.id = record.id.split("_")[0]
                    busco_ids_retrieved.add(record.id)
                    matched_seqs.append(record)

        unmatched_incomplete_buscos = list(
            set(self.incomplete_buscos) - set(busco_ids_retrieved)
        )
        if len(unmatched_incomplete_buscos) > 0:
            logger.debug(
                "The BUSCO ID(s) {} were not found in the file {}".format(
                    unmatched_incomplete_buscos, os.path.basename(self.refseq_db)
                )
            )

        with open(
            self.refseq_db_rerun, "w"
        ) as out_file:  # Create new query file for second tblastn run
            SeqIO.write(matched_seqs, out_file, "fasta")

        return

    def get_version(self):
        help_output = subprocess.check_output(
            [self.cmd, "-h"], stderr=subprocess.STDOUT, shell=False
        )
        lines = help_output.decode("utf-8").split("\n")
        version = None
        for line in lines:
            if line.startswith("metaeuk Version:"):
                version = line.strip().split(" ")[-1]
        return version

    def parse_parameters(self):
        accepted_keys = []
        accepted_values = []
        if self.extra_params:
            self.extra_params = self.extra_params.strip("\" '")
            try:
                if self.extra_params.startswith("-"):
                    key_val_pairs = self.extra_params.split(" -")
                    for kv in key_val_pairs:
                        key_vals = kv.strip("- ").split("=")
                        if len(key_vals) == 2:
                            key, val = key_vals
                            if key in type(self).ACCEPTED_PARAMETERS:
                                if key == "min-exon-aa":
                                    self.min_exon_aa = val.strip()
                                    continue
                                elif key == "max-intron":
                                    self.max_intron = val.strip()
                                    continue
                                elif key == "max-seq-len":
                                    self.max_seq_len = val.strip()
                                    continue
                                elif key == "max-overlap":
                                    self.max_overlap = val.strip()
                                    continue
                                elif key == "min-intron":
                                    self.min_intron = val.strip()
                                    continue
                                elif key == "overlap":
                                    self.overlap = val.strip()
                                    continue
                                elif key == "s":
                                    self.s_set = True
                                accepted_keys.append(key.strip())
                                accepted_values.append(val.strip())
                            else:
                                logger.warning(
                                    "{} is not an accepted parameter for Metaeuk.".format(
                                        key
                                    )
                                )
                        else:
                            raise MetaeukParsingError
                else:
                    raise MetaeukParsingError
            except MetaeukParsingError:
                logger.warning(
                    "Metaeuk parameters are not correctly formatted. Please enter them as follows: "
                    '"--param1=value1 --param2=value2" etc. Proceeding without additional parameters.'
                )
                return [], []
        return accepted_keys, accepted_values

    @staticmethod
    def parse_header(header):
        header_parts = header.split("|")
        if not header_parts[2] in [
            "+",
            "-",
        ]:  # Deal with sequence IDs that contain the symbol "|"
            try:
                strand_ind = header_parts.index("+")
            except ValueError:
                strand_ind = header_parts.index("-")
            header_parts[1] = "|".join(header_parts[1:strand_ind])
            for i in range(2, strand_ind):
                header_parts.pop(i)

        T_acc = header_parts[0]
        C_acc = header_parts[1]
        strand = header_parts[2]
        bitscore = float(header_parts[3])
        eval = float(header_parts[4])
        num_exons = int(header_parts[5])
        low_coord = int(header_parts[6])
        high_coord = int(header_parts[7])
        exon_coords = header_parts[8:]

        all_low_exon_coords = []
        all_taken_low_exon_coords = []
        all_high_exon_coords = []
        all_taken_high_exon_coords = []
        all_exon_nucl_len = []
        all_taken_exon_nucl_len = []
        for exon in exon_coords:
            low_exon_coords, high_exon_coords, nucl_lens = exon.split(":")

            low_exon_coord, taken_low_exon_coord = low_exon_coords.split("[")
            all_low_exon_coords.append(int(low_exon_coord))
            taken_low_exon_coord = int(taken_low_exon_coord.strip("]"))

            high_exon_coord, taken_high_exon_coord = high_exon_coords.split("[")
            all_high_exon_coords.append(int(high_exon_coord))
            taken_high_exon_coord = int(taken_high_exon_coord.strip("]"))

            nucl_len, taken_nucl_len = nucl_lens.split("[")
            all_exon_nucl_len.append(int(nucl_len))
            taken_nucl_len = int(taken_nucl_len.strip().rstrip("]"))

            # Need to fix the metaeuk coordinate problem
            if strand == "-":
                if int(taken_high_exon_coord) + int(taken_nucl_len) - 1 != int(
                    taken_low_exon_coord
                ):
                    taken_low_exon_coord = (
                        int(taken_high_exon_coord) + int(taken_nucl_len) - 1
                    )

            all_taken_low_exon_coords.append(taken_low_exon_coord)
            all_taken_high_exon_coords.append(taken_high_exon_coord)
            all_taken_exon_nucl_len.append(taken_nucl_len)

        gene_id = "{}:{}-{}".format(C_acc, low_coord, high_coord)

        details = {
            "T_acc": T_acc,
            "C_acc": C_acc,
            "S": strand,
            "bitscore": bitscore,
            "e-value": eval,
            "num_exons": num_exons,
            "low_coord": low_coord,
            "high_coord": high_coord,
            "all_low_exon_coords": all_low_exon_coords,
            "all_taken_low_exon_coords": all_taken_low_exon_coords,
            "all_high_exon_coords": all_high_exon_coords,
            "all_taken_high_exon_coords": all_taken_high_exon_coords,
            "all_exon_nucl_len": all_exon_nucl_len,
            "all_taken_exon_nucl_len": all_taken_exon_nucl_len,
            "gene_id": gene_id,
        }
        return details

    def edit_protein_file(self):
        all_records = []
        all_headers = []
        try:
            with open(self.pred_protein_seqs, "rU") as f:
                for record in SeqIO.parse(f, "fasta"):
                    header_details = self.parse_header(record.id)
                    record.id = header_details["gene_id"]
                    record.name = header_details["gene_id"]
                    record.description = header_details["gene_id"]
                    all_records.append(record)

                    header_df_info = (
                        header_details["T_acc"].split("_")[0],
                        header_details["C_acc"],
                        header_details["low_coord"],
                        header_details["high_coord"],
                        header_details["S"],
                        header_details["bitscore"],
                        None,
                    )
                    all_headers.append(header_df_info)

        except FileNotFoundError:
            raise NoGenesError("Metaeuk")

        if all_headers:
            matches_df = self.records_to_df(all_headers)
            overlaps = self.test_for_overlaps(matches_df)
            inds_to_remove = []
            for overlap in overlaps:
                match1 = matches_df.loc[overlap[0]]
                match2 = matches_df.loc[overlap[1]]
                if match1["Busco id"] == match2["Busco id"]:
                    if float(match1["Score"]) > float(match2["Score"]):
                        ind_to_remove = match2.name
                    else:
                        ind_to_remove = match1.name
                    inds_to_remove.append(ind_to_remove)

            filtered_records = [
                i for j, i in enumerate(all_records) if j not in inds_to_remove
            ]
            for record in filtered_records:
                self.sequences_aa[record.id] = record

            with open(self.pred_protein_seqs_modified, "w") as f_mod:
                SeqIO.write(filtered_records, f_mod, "fasta")

    def get_gene_details(self):
        self.gene_details = defaultdict(list)

        try:
            with open(self.headers_file, "r") as f:
                lines = f.readlines()

            try:
                for line in lines:
                    header = line.split("\t")[-1]
                    header_details = self.parse_header(header)

                    self.gene_details[header_details["gene_id"]].append(
                        {
                            "gene_start": header_details["low_coord"],
                            "gene_end": header_details["high_coord"],
                            "strand": header_details["S"],
                        }
                    )

            except KeyError:
                raise SystemExit("*headersMap.tsv file could not be parsed.")
        except FileNotFoundError:
            raise NoRerunFile



#!/usr/bin/env python3
# coding: utf-8

"""
.. module:: BuscoPlacer
   :synopsis: BuscoPlacer implements methods required for automatically selecting the appropriate dataset
   to be used during BUSCO analysis
.. versionadded:: 4.0.0
.. versionchanged:: 4.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""
import json
import os
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from Bio import SeqIO
from busco.busco_tools.sepp import SEPPRunner

logger = BuscoLogger.get_logger(__name__)


class BuscoPlacer:

    _logger = BuscoLogger.get_logger(__name__)

    @log(
        "***** Searching tree for chosen lineage to find best taxonomic match *****\n",
        logger,
    )
    def __init__(self, config, run_folder, protein_seqs, single_copy_buscos):
        self._config = config
        self._params = config
        self.mode = self._config.get("busco_run", "mode")
        self.cpus = self._config.get("busco_run", "cpu")
        self.restart = self._config.getboolean("busco_run", "restart")
        self.run_folder = run_folder
        self.placement_folder = os.path.join(run_folder, "placement_files")
        if self.restart:
            os.makedirs(self.placement_folder, exist_ok=True)
        else:
            os.mkdir(self.placement_folder)
        self.downloader = self._config.downloader
        self.datasets_version = self._config.get("busco_run", "datasets_version")
        self.protein_seqs = protein_seqs
        self.single_copy_buscos = single_copy_buscos  # dict
        self.init_tools()

    def _download_placement_files(self):
        self.ref_markers_file = self.downloader.get(
            "list_of_reference_markers.{0}_{1}.txt".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        self.tree_nwk_file = self.downloader.get(
            "tree.{0}_{1}.nwk".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        self.tree_metadata_file = self.downloader.get(
            "tree_metadata.{0}_{1}.txt".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        self.supermatrix_file = self.downloader.get(
            "supermatrix.aln.{0}_{1}.faa".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        self.taxid_busco_file = self.downloader.get(
            "mapping_taxids-busco_dataset_name.{0}_{1}.txt".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        self.taxid_lineage_file = self.downloader.get(
            "mapping_taxid-lineage.{0}_{1}.txt".format(
                os.path.basename(self.run_folder).split("_")[-2], self.datasets_version
            ),
            "placement_files",
        )
        return

    def _get_placement_file_versions(self):
        placement_file_versions = [
            os.path.basename(filepath)
            for filepath in [
                self.ref_markers_file,
                self.tree_nwk_file,
                self.tree_metadata_file,
                self.supermatrix_file,
                self.taxid_busco_file,
                self.taxid_lineage_file,
            ]
        ]
        return placement_file_versions

    @log("Extract markers...", logger)
    def define_dataset(self):
        # If mode is genome, substitute input with prodigal/augustus output
        self._download_placement_files()
        placement_file_versions = self._get_placement_file_versions()
        self._extract_marker_sequences()
        self._run_sepp()

        dataset = self._pick_dataset()

        return dataset, placement_file_versions

    def init_tools(self):
        setattr(SEPPRunner, "config", self._config)
        self.sepp_runner = SEPPRunner()

    def _pick_dataset(self):

        run_folder = self.run_folder

        # load busco dataset name by id in a dict {taxid:name}
        datasets_mapping = {}

        with open(self.taxid_busco_file) as f:
            for line in f:
                datasets_mapping.update(
                    {
                        line.strip()
                        .split("\t")[0]: line.strip()
                        .split("\t")[1]
                        .split(",")[0]
                    }
                )

        # load the lineage for each taxid in a dict {taxid:reversed_lineage}
        # lineage is 1:2:3:4:5:6 => {6:[6,5,4,3,2,1]}
        lineages = set()
        parents = {}
        taxid_dataset = {}
        for t in datasets_mapping:
            taxid_dataset.update({t: t})
        with open(self.taxid_lineage_file) as f:
            for line in f:
                if line.startswith("#"):
                    continue
                lineages.add(line.strip().split("\t")[4])

                i = 0
                # for each line, e.g. 6\t1:2:3:4:5:6, create/update the lineage for each level
                # 6:[1,2,3,4,5,6], 5:[1,2,3,4,5], 4:[1,2,3,4], etc.
                for t in line.strip().split("\t")[4].split(","):
                    i += 1
                    parents.update({t: line.strip().split("\t")[4].split(",")[0:i]})

        for t in parents:
            for p in parents[t][
                ::-1
            ]:  # reverse the order to get the deepest parent, not the root one
                if p in datasets_mapping:
                    taxid_dataset.update({t: p})
                    break
        # load json
        # load "tree" in a string
        # load placements
        # obtain a dict of taxid num of markers
        # figure out which taxid to use by using the highest number of markers and some extra rules

        try:
            with open(
                os.path.join(self.placement_folder, "output_placement.json")
            ) as json_file:
                data = json.load(json_file)
            tree = data["tree"]
            placements = data["placements"]
        except FileNotFoundError:
            raise SystemExit(
                "Placements failed. Try to rerun increasing the memory or select a lineage manually."
            )

        node_weight = {}
        n_p = 0
        for placement in placements:
            n_p += 1
            for individual_placement in placement["p"]:
                # find the taxid in tree
                node = individual_placement[0]
                import re

                match = re.findall(  # deal with weird character in the json file, see the output yourself.
                    # if this pattern is inconsistant with pplacer version, it may break buscoplacer.
                    "[^0-9][0-9]*:[0-9]*[^0-9]{0,1}[0-9]*[^0-9]{0,2}[0-9]*\[%s\]"
                    % node,
                    tree,
                )
                # extract taxid:
                try:
                    if re.match("^[A-Za-z]", match[0]):
                        taxid = match[0][7:].split(":")[0]
                    else:
                        taxid = match[0][1:].split(":")[0]
                except IndexError as e:
                    raise e
                if taxid_dataset[taxid] in node_weight:
                    node_weight[taxid_dataset[taxid]] += 1
                else:
                    node_weight[taxid_dataset[taxid]] = 1
                break  # Break here to keep only the best match. In my experience, keeping all does not change much.
        type(self)._logger.debug("Placements counts by node are: %s" % node_weight)

        # from here, define which placement can be trusted
        max_markers = 0
        choice = []

        for key in node_weight:
            type(self)._logger.debug(
                "%s markers assigned to the taxid %s" % (node_weight[key], key)
            )

        # taxid for which no threshold or minimal amount of placement should be considered.
        # If it is the best, go for it.
        no_rules = ["204428"]

        ratio = 2.5
        if run_folder.split("/")[-1].split("_")[-2] == "archaea":
            ratio = 1.2
        min_markers = 12

        node_with_max_markers = None
        for n in node_weight:
            if node_weight[n] > max_markers:
                max_markers = node_weight[n]
                node_with_max_markers = n
        if node_with_max_markers in no_rules:
            choice = [node_with_max_markers]
        else:
            for n in node_weight:
                # if the ration between the best and the current one is not enough, keep both
                if node_weight[n] * ratio >= max_markers:
                    choice.append(n)
        if len(choice) > 1:
            # more than one taxid should be considered, pick the common ancestor
            choice = self._get_common_ancestor(choice, parents)
            # print('last common')
            # print(choice)
        elif len(choice) == 0:
            if run_folder.split("/")[-1].split("_")[-2] == "bacteria":
                choice.append("2")
            elif run_folder.split("/")[-1].split("_")[-2] == "archaea":
                choice.append("2157")
            elif run_folder.split("/")[-1].split("_")[-2] == "eukaryota":
                choice.append("2759")
        if max_markers < min_markers and not (choice[0] in no_rules):
            if run_folder.split("/")[-1].split("_")[-2] == "bacteria":
                key_taxid = "2"
            elif run_folder.split("/")[-1].split("_")[-2] == "archaea":
                key_taxid = "2157"
            elif run_folder.split("/")[-1].split("_")[-2] == "eukaryota":
                key_taxid = "2759"
            else:
                key_taxid = None  # unexpected. Should throw an exception or use assert.
            type(self)._logger.info(
                "Not enough markers were placed on the tree (%s). Root lineage %s is kept"
                % (max_markers, datasets_mapping[taxid_dataset[key_taxid]])
            )
            return [
                datasets_mapping[taxid_dataset[key_taxid]],
                max_markers,
                sum(node_weight.values()),
            ]

        type(self)._logger.info(
            "Lineage %s is selected, supported by %s markers out of %s"
            % (
                datasets_mapping[taxid_dataset[choice[0]]],
                max_markers,
                sum(node_weight.values()),
            )
        )

        return [
            datasets_mapping[taxid_dataset[choice[0]]],
            max_markers,
            sum(node_weight.values()),
        ]

    @staticmethod
    def _get_common_ancestor(choice, parents):
        # starts with the parents of the first choice
        all_ancestors = set(parents[choice[0]])
        # order will be lost with sets, so keep in a list the lineage of one entry to later pick the deepest ancestor
        ordered_lineage = []
        for c in choice:
            # print('c is %s' % c)
            if len(parents[c]) > len(ordered_lineage):
                #    print('len parent c is %s' % len(parents[c]))
                #    print('len ordered lineage is %s' % ordered_lineage)
                # probably useless. Init with parents[choice[0] should work
                ordered_lineage = parents[c]
            #    print('ordered_lineage us %s' % ordered_lineage)
            # keep in set only entries that are in the currently explored lineage
            all_ancestors = all_ancestors.intersection(parents[c])

        # go through the ordered list of the deepest linage until you found a common ancestor.
        for parent in ordered_lineage[::-1]:
            if parent in all_ancestors:
                return [parent]

    @log("Place the markers on the reference tree...", logger)
    def _run_sepp(self):
        self.sepp_runner.configure_runner(
            self.tree_nwk_file,
            self.tree_metadata_file,
            self.supermatrix_file,
            self.downloader,
        )
        if self.restart and self.sepp_runner.check_previous_completed_run():
            logger.info("Skipping SEPP run as it has already been completed")
        else:
            self.restart = False
            self._config.set("busco_run", "restart", str(self.restart))
            self.sepp_runner.run()
            self.sepp_runner.cleanup()

    def _extract_marker_sequences(self):
        """
        This function extracts all single copy BUSCO genes from a protein run folder
        :param run_folder: a BUSCO protein run folder
        :type: str
        :param protein_sequences: protein fasta used as input for the BUSCO run corresponding to the folder
        :type: str
        """

        with open(self.ref_markers_file, "r") as f:
            marker_list = [line.strip() for line in f]

        marker_genes_names = []
        for busco, gene_matches in self.single_copy_buscos.items():
            if busco in marker_list:
                marker_genes_names.append(
                    list(gene_matches.keys())[0]
                )  # The list should only have one entry because they are single copy buscos

        marker_genes_records = []
        if isinstance(self.protein_seqs, (str,)):
            list_protein_seqs = [self.protein_seqs]
        else:
            list_protein_seqs = self.protein_seqs

        for protein_seqs in list_protein_seqs:
            with open(protein_seqs, "r") as prot_seqs:
                for record in SeqIO.parse(prot_seqs, "fasta"):
                    if record.id in marker_genes_names:
                        record.seq = record.seq.rstrip("*")
                        record.description = ""
                        marker_genes_records.append(record)

        marker_genes_file = os.path.join(self.placement_folder, "marker_genes.fasta")
        with open(marker_genes_file, "w") as output:
            SeqIO.write(marker_genes_records, output, "fasta")



from busco.busco_tools.base import BaseRunner
import os
from busco.BuscoLogger import BuscoLogger
import shutil
import subprocess

logger = BuscoLogger.get_logger(__name__)


class SEPPRunner(BaseRunner):

    name = "sepp"
    cmd = "run_sepp.py"

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(
            self.main_out, "auto_lineage", self.lineage_results_dir
        )
        self.placement_folder = os.path.join(self._output_folder, "placement_files")
        self._tmp_folder = os.path.join(self._output_folder, "sepp_tmp_files")
        self.datasets_version = self.config.get("busco_run", "datasets_version")
        self.create_dirs([self._output_folder, self.placement_folder, self._tmp_folder])

        self.init_checkpoint_file()

    def configure_runner(
        self, tree_nwk_file, tree_metadata_file, supermatrix_file, downloader
    ):
        self.run_number += 1
        self.tree_nwk_file = tree_nwk_file
        self.tree_metadata_file = tree_metadata_file
        self.supermatrix_file = supermatrix_file
        self.downloader = downloader

    def cleanup(self):
        shutil.rmtree(self._tmp_folder)

    def generate_job_args(self):
        yield

    def run(self):
        super().run()
        self.total = 1
        self.run_jobs()

    def configure_job(self, *args):
        sepp_job = self.create_job()
        sepp_job.add_parameter("--cpu")
        sepp_job.add_parameter(str(self.cpus))
        sepp_job.add_parameter("--outdir")
        sepp_job.add_parameter(self.placement_folder)
        sepp_job.add_parameter("-t")
        sepp_job.add_parameter(self.tree_nwk_file)
        sepp_job.add_parameter("-r")
        sepp_job.add_parameter(self.tree_metadata_file)
        sepp_job.add_parameter("-a")
        sepp_job.add_parameter(self.supermatrix_file)
        sepp_job.add_parameter("-f")
        sepp_job.add_parameter(
            os.path.join(self.placement_folder, "marker_genes.fasta")
        )
        sepp_job.add_parameter("-F")
        sepp_job.add_parameter("15")
        sepp_job.add_parameter("-m")
        sepp_job.add_parameter("amino")
        sepp_job.add_parameter("-p")
        sepp_job.add_parameter(self._tmp_folder)
        return sepp_job

    def check_tool_dependencies(self):
        pass

    def get_version(self):
        sepp_version = subprocess.check_output(
            [self.cmd, "-v"], stderr=subprocess.STDOUT, shell=False
        )
        sepp_version = sepp_version.decode("utf-8")
        sepp_version = sepp_version.strip().split(" ")[1]
        return sepp_version

    @property
    def output_folder(self):
        return self._output_folder



#!/usr/bin/env python3
# coding: utf-8
"""
.. module:: TranscriptomeAnalysis
   :synopsis:TranscriptomeAnalysis implements genome analysis specifics
.. versionadded:: 3.0.0
.. versionchanged:: 5.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""
import os
from busco.analysis.BuscoAnalysis import BuscoAnalysis
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from Bio.Seq import reverse_complement, translate
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from busco.analysis.Analysis import NucleotideAnalysis, BLASTAnalysis
from busco.analysis.GenomeAnalysis import GenomeAnalysisEukaryotesMetaeuk
from abc import ABCMeta
from collections import defaultdict


logger = BuscoLogger.get_logger(__name__)


class TranscriptomeAnalysis(NucleotideAnalysis, metaclass=ABCMeta):
    """
    Analysis on a transcriptome.
    """

    _mode = "transcriptome"

    def __init__(self):
        """
        Initialize an instance.
        """
        super().__init__()


class TranscriptomeAnalysisProkaryotes(
    TranscriptomeAnalysis, BLASTAnalysis, BuscoAnalysis
):
    """
    Analysis on a transcriptome.
    """

    def __init__(self):
        """
        Initialize an instance.
        """
        super().__init__()
        self.all_sequences = defaultdict(lambda: defaultdict(dict))

        self.sequences_aa = {}
        self.complete_seqs_nt = {}
        self.complete_seqs_aa = {}

        self.single_copy_proteins_file = os.path.join(
            self.run_folder, "single_copy_proteins.faa"
        )

    def run_analysis(self):
        """
        This function calls all needed steps for running the analysis.
        """

        super().run_analysis()

        self._run_mkblast()
        self._run_tblastn(ancestral_variants=self._has_variants_file)

        protein_seq_files = self._translate_seqs(self.tblastn_runner.coords)

        self.run_hmmer(protein_seq_files)

        self.prepare_sequences()
        self.write_complete_seqs()

        self.hmmer_runner.write_buscos_to_file(self.sequences_aa, self.sequences_nt)

        # if self._tarzip:
        #     self._run_tarzip_hmmer_output()
        #     self._run_tarzip_translated_proteins()
        return

    def init_tools(self):
        super().init_tools()

    def cleanup(self):
        """
        This function cleans temporary files.
        """
        super().cleanup()

    def match_hmmer_results(self, hmmer_results):
        seqs_aa = {}
        seqs_nt = {}

        for busco_id, gene_matches in hmmer_results.items():
            for gene_id in gene_matches:
                gene_info = gene_matches[gene_id][0]
                frame = gene_info["frame"]
                seq_aa = self.all_sequences[busco_id][gene_id]["translations"][frame]
                seq_nt = self.all_sequences[busco_id][gene_id]["original"]
                seqs_aa[gene_id] = seq_aa
                seqs_nt[gene_id] = seq_nt
        return seqs_aa, seqs_nt

    def prepare_sequences(self):
        hmmer_results_complete = self.hmmer_runner.single_copy_buscos

        sc_seqs_aa, sc_seqs_nt = self.match_hmmer_results(hmmer_results_complete)
        self.complete_seqs_aa = [val for key, val in sc_seqs_aa.items()]

        hmmer_results_remainder = {
            **self.hmmer_runner.multi_copy_buscos,
            **self.hmmer_runner.fragmented_buscos,
        }

        remainder_seqs_aa, remainder_seqs_nt = self.match_hmmer_results(
            hmmer_results_remainder
        )

        self.sequences_aa = {**sc_seqs_aa, **remainder_seqs_aa}
        self.sequences_nt = {**sc_seqs_nt, **remainder_seqs_nt}

        return

    def write_complete_seqs(self):

        with open(self.single_copy_proteins_file, "w") as out_faa:
            SeqIO.write(self.complete_seqs_aa, out_faa, "fasta")

        return

    @staticmethod
    def six_frame_translation(record):
        """
        Gets the sixframe translation for the provided sequence
        :param seq: the sequence to be translated
        :type seq: str
        :return: the six translated sequences
        :rtype: list
        """
        descriptions = {
            1: "orig_seq_frame_1",
            2: "orig_seq_frame_2",
            3: "orig_seq_frame_3",
            -1: "rev_comp_frame_1",
            -2: "rev_comp_frame_2",
            -3: "rev_comp_frame_3",
        }

        # Based on code excerpt from https://biopython.org/DIST/docs/api/Bio.SeqUtils-pysrc.html#six_frame_translations
        anti = reverse_complement(record.seq)
        translated_seqs = {}
        for i in range(3):
            fragment_length = 3 * ((len(record.seq) - i) // 3)
            translated_seqs[descriptions[i + 1]] = SeqRecord(
                translate(record.seq[i : i + fragment_length], stop_symbol="X"),
                id=record.id,
                name=record.id,
                description=descriptions[i + 1],
            )
            translated_seqs[descriptions[-(i + 1)]] = SeqRecord(
                translate(anti[i : i + fragment_length], stop_symbol="X"),
                id=record.id,
                name=record.id,
                description=descriptions[i + 1],
            )
        return translated_seqs

    @staticmethod
    def _reformats_seq_id(seq_id):
        """
        This function reformats the sequence id to its original values
        :param seq_id: the seq id to reformats
        :type seq_id: str
        :return: the reformatted seq_id
        :rtype: str
        """
        return "_".join(seq_id.split("_")[:-1])

    @log("Translating candidate transcripts", logger)
    def _translate_seqs(self, coords):

        translated_proteins_dir = os.path.join(self.main_out, "translated_proteins")
        if not os.path.exists(translated_proteins_dir):
            os.makedirs(translated_proteins_dir)

        contig_names = []
        for contig_info in coords.values():
            for contig in contig_info:
                contig_names.append(contig)

        protein_seq_files = []
        for busco_id, contig_info in coords.items():
            output_filename = os.path.join(
                translated_proteins_dir, "{}.faa".format(busco_id)
            )
            protein_seq_files.append(output_filename)
            translated_records = []
            for contig_name in contig_info:
                tmp_filename = os.path.join(
                    self.tblastn_runner.output_seqs, "{}.temp".format(contig_name[:100])
                )  # Avoid very long filenames
                for record in SeqIO.parse(
                    tmp_filename, "fasta"
                ):  # These files will only ever have one sequence,
                    # but BioPython examples always parse them in an iterator.
                    translated_seqs = self.six_frame_translation(record)
                    gene_id = "{}:{}-{}".format(
                        record.id,
                        contig_info[contig_name]["contig_start"],
                        contig_info[contig_name]["contig_end"],
                    )
                    self.all_sequences[busco_id][gene_id][
                        "translations"
                    ] = translated_seqs
                    self.all_sequences[busco_id][gene_id]["original"] = record
                    for (
                        desc_id
                    ) in translated_seqs:  # There are six possible translated sequences
                        prot_seq = translated_seqs[desc_id]
                        prot_seq.id = gene_id
                        translated_records.append(prot_seq)

            with open(output_filename, "w") as out_faa:
                SeqIO.write(translated_records, out_faa, "fasta")

        return protein_seq_files

    # def _run_tarzip_translated_proteins(self):
    #     """
    #     This function tarzips results folder
    #     """
    #     # translated_proteins # Todo: rewrite with tarfile module
    #     self._p_open(["tar", "-C", "%s" % self.mainout, "-zcf",
    #                  "%stranslated_proteins.tar.gz" % self.mainout, "translated_proteins", "--remove-files"], "bash",
    #                  shell=False)


class TranscriptomeAnalysisEukaryotes(
    TranscriptomeAnalysis, GenomeAnalysisEukaryotesMetaeuk
):
    def __init__(self):
        super().__init__()



import unittest
from busco import BuscoConfig
import shutil
import os
from unittest.mock import Mock
from unittest.mock import patch
from pathlib import Path


class TestBuscoConfig(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.maxDiff = None
        self.base_config = "config/config.ini"

        self.params = {
            "auto-lineage": False,
            "auto-lineage-euk": False,
            "auto-lineage-prok": False,
            "config_file": None,
            "cpu": None,
            "evalue": None,
            "force": False,
            "help": "==SUPPRESS==",
            "in": None,
            "limit": None,
            "lineage_dataset": None,
            "list_datasets": "==SUPPRESS==",
            "mode": None,
            "offline": False,
            "out": None,
            "out_path": None,
            "quiet": False,
            "restart": False,
            "metaeuk_parameters": None,
            "metaeuk_rerun_parameters": None,
            "use_augustus": False,
            "augustus_parameters": None,
            "augustus_species": None,
            "long": False,
            "datasets_version": None,
            "download_base_url": None,
            "download_path": None,
            "update-data": False,
            "version": "==SUPPRESS==",
        }

        self.test_params = {
            "in": "input_test",
            "out": "output_test",
            "mode": "mode_test",
        }

        self.config_structure = {
            "augustus": ["path", "command"],
            "busco_run": [
                "in",
                "out",
                "out_path",
                "mode",
                "auto-lineage",
                "auto-lineage-prok",
                "auto-lineage-euk",
                "cpu",
                "force",
                "restart",
                "download_path",
                "datasets_version",
                "quiet",
                "offline",
                "long",
                "augustus_parameters",
                "augustus_species",
                "download_base_url",
                "lineage_dataset",
                "update-data",
                "metaeuk_parameters",
                "metaeuk_rerun_parameters",
                "evalue",
                "limit",
                "use_augustus",
            ],
            "etraining": ["path", "command"],
            "gff2gbSmallDNA.pl": ["path", "command"],
            "hmmsearch": ["path", "command"],
            "makeblastdb": ["path", "command"],
            "metaeuk": ["path", "command"],
            "new_species.pl": ["path", "command"],
            "optimize_augustus.pl": ["path", "command"],
            "prodigal": ["path", "command"],
            "sepp": ["path", "command"],
            "tblastn": ["path", "command"],
        }

    def test_read_config_file(self):
        config = BuscoConfig.BaseConfig()
        config.conf_file = self.base_config
        config._load_config_file()
        self.assertIn("busco_run", config.sections())

    def test_read_config_file_ioerror(self):
        with self.assertRaises(SystemExit):
            config = BuscoConfig.BaseConfig()
            config.conf_file = "/path/not/found"
            config._load_config_file()

    def test_read_config_file_parseerror(self):
        config_path = "tests/config_parseerror_test.ini"
        test_config_contents = "in=input_file\n"
        with open(config_path, "w") as f:
            f.write(test_config_contents)

        with self.assertRaises(SystemExit):
            config = BuscoConfig.BaseConfig()
            config.conf_file = config_path
            config._load_config_file()
        os.remove(config_path)

    def test_read_config_file_duplicateerror(self):
        config_path = "tests/config_duplicate_test.ini"
        test_config_contents = "[busco_run]\n" "in=input_file\n" "in=input_file\n"
        with open(config_path, "w") as f:
            f.write(test_config_contents)

        with self.assertRaises(SystemExit):
            config = BuscoConfig.BaseConfig()
            config.conf_file = config_path
            config._load_config_file()
        os.remove(config_path)

    def test_config_update_args_bool(self):
        update_params = {
            "force": True,
            "offline": True,
            "quiet": True,
            "restart": True,
        }
        config = BuscoConfig.BuscoConfigMain(self.base_config, update_params)
        config.configure()
        self.assertEqual(
            update_params,
            {key: config.getboolean("busco_run", key) for key in update_params.keys()},
        )

    def test_config_update_args_nonbool(self):
        update_params = {
            "cpu": "10",
            "evalue": "0.01",
            "in": "input_file",
            "limit": "1",
            "lineage_dataset": "test",
            "mode": "test",
            "out": "test",
            "out_path": "test",
        }
        config = BuscoConfig.BuscoConfigMain(self.base_config, update_params)
        config.configure()
        self.assertEqual(
            update_params,
            {key: config.get("busco_run", key) for key in update_params.keys()},
        )

    def test_config_default_params(self):
        correct_default_params = {
            "auto-lineage": False,
            "auto-lineage-euk": False,
            "auto-lineage-prok": False,
            "cpu": "1",
            "datasets_version": "odb10",
            "download_base_url": "https://busco-data.ezlab.org/v5/data/",
            "download_path": os.path.join(os.getcwd(), "busco_downloads"),
            "evalue": "0.001",
            "force": False,
            "limit": "3",
            "long": False,
            "offline": False,
            "out_path": os.getcwd(),
            "quiet": False,
            "restart": False,
            "update-data": False,
            "use_augustus": False,
        }
        config = BuscoConfig.BuscoConfigMain(self.base_config, {})
        config.configure()
        config_default_filled = {
            key: config.get("busco_run", key) for key in correct_default_params
        }

        self.assertEqual(
            {key: str(val) for key, val in correct_default_params.items()},
            config_default_filled,
        )

    @patch(
        "busco.BuscoConfig.BuscoConfigMain.getboolean",
        side_effect=[True, False, False, True, False],
    )
    def test_config_auto_lineage_settings(self, *args):
        for _ in range(2):
            config = BuscoConfig.BuscoConfigMain(self.base_config, {})
            config.configure()
            self.assertEqual(config.get("busco_run", "auto-lineage"), "True")

    @patch("busco.BuscoConfig.BuscoConfigMain.getboolean", return_value=True)
    def test_config_auto_lineage_both_selected_warning(self, *args):
        with self.assertLogs(BuscoConfig.logger, "WARNING"):
            config = BuscoConfig.BuscoConfigMain(self.base_config, {})
            config.configure()
        self.assertEqual(config.get("busco_run", "auto-lineage-euk"), "False")
        self.assertEqual(config.get("busco_run", "auto-lineage-prok"), "False")

    def test_mandatory_keys_check_log(self):
        with self.assertLogs(BuscoConfig.logger, 20):
            params_test = {"in": "input_file", "out": "output_name", "mode": "genome"}
            config = BuscoConfig.BuscoConfigMain(self.base_config, params_test)
            config.configure()
            config._check_mandatory_keys_exist()

    def test_mandatory_keys_check_missing_param_in(self):
        with self.assertRaises(SystemExit):
            params_test = {"out": "output_name", "mode": "genome"}
            config = BuscoConfig.BuscoConfigMain(self.base_config, params_test)
            config.configure()
            config._check_mandatory_keys_exist()

    def test_mandatory_keys_check_missing_param_mode(self):
        with self.assertRaises(SystemExit):
            params_test = {"in": "input_file", "out": "output_name"}
            config = BuscoConfig.BuscoConfigMain(self.base_config, params_test)
            config.configure()
            config._check_mandatory_keys_exist()

    def test_mandatory_keys_check_missing_param_out(self):
        with self.assertRaises(SystemExit):
            params_test = {"in": "input_file", "mode": "genome"}
            config = BuscoConfig.BuscoConfigMain(self.base_config, params_test)
            config.configure()
            config._check_mandatory_keys_exist()

    def test_previous_run_check_without_existing_run(self):
        output_dir = os.path.join(os.getcwd(), self.test_params["out"])
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        self.assertIsNone(config._check_no_previous_run())

    def test_previous_run_check_with_existing_run_no_force(self):
        previous_run_name = "test_busco_run_dir"
        os.makedirs(previous_run_name, exist_ok=True)
        self.test_params["out"] = previous_run_name
        self.test_params["force"] = "False"
        with self.assertRaises(SystemExit):
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            config._check_no_previous_run()
        shutil.rmtree(previous_run_name)

    def test_previous_run_check_with_existing_run_with_force_and_log(self):
        previous_run_name = "test_busco_run_dir"
        os.makedirs(previous_run_name, exist_ok=True)
        self.test_params["out"] = previous_run_name
        self.test_params["force"] = "True"
        with self.assertLogs(BuscoConfig.logger, 20):
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            config._check_no_previous_run()
            self.assertFalse(os.path.exists(previous_run_name))

        try:  # In case of test failure, remove tmp folder anyway
            shutil.rmtree(previous_run_name)
        except FileNotFoundError:
            pass

    def test_previous_run_check_without_existing_run_and_restart(self):
        self.test_params["restart"] = "True"
        with self.assertLogs(BuscoConfig.logger, "WARNING"):
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            config._check_no_previous_run()
            self.assertEqual(config.getboolean("busco_run", "restart"), False)

    def test_previous_run_check_with_existing_run_and_restart(self):
        previous_run_name = "test_busco_run_dir"
        os.makedirs(previous_run_name, exist_ok=True)
        self.test_params.update({"out": previous_run_name, "restart": True})
        with self.assertLogs(BuscoConfig.logger, "INFO"):
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            config._check_no_previous_run()
            self.assertEqual(config.getboolean("busco_run", "restart"), True)
        shutil.rmtree(previous_run_name)

    def test_create_required_paths(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.main_out = os.path.join(
            config.get("busco_run", "out_path"), config.get("busco_run", "out")
        )
        config._create_required_paths()
        output_dir = os.path.join(os.getcwd(), self.test_params["out"])
        self.assertTrue(os.path.exists(output_dir))
        shutil.rmtree(output_dir)

    def test_config_structure(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        self.assertEqual(
            set(config.PERMITTED_OPTIONS), set(self.config_structure["busco_run"])
        )

    def test_catch_disallowed_keys(self):
        for section_name in self.config_structure:
            with self.assertRaises(SystemExit):
                config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
                config.configure()
                config.set(section_name, "forbidden_option", "forbidden_value")
                config._check_allowed_keys()

    def test_out_value_check_invalid(self):
        for str_format in ["/path/to/output", "output/"]:
            self.test_params["out"] = str_format
            with self.assertRaises(SystemExit):
                config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
                config.configure()
                config._check_out_value()

    def test_out_value_check_valid(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        self.assertIsNone(config._check_out_value())

    def test_limit_value_out_of_range(self):
        for lim_val in [-1, 0, 25]:
            self.test_params["limit"] = lim_val
            with self.assertRaises(SystemExit):
                config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
                config.configure()
                config._check_limit_value()

    def test_limit_value_within_range(self):
        for lim_val in [1, 20]:
            self.test_params["limit"] = lim_val
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            self.assertIsNone(config._check_limit_value())

    def test_evalue_nondefault(self):
        self.test_params["evalue"] = 1
        with self.assertLogs(BuscoConfig.logger, level="WARNING"):
            config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
            config.configure()
            config._check_evalue()

    @patch("__main__.BuscoConfig_unittests.BuscoConfig.logger.warning")
    def test_evalue_default(self, mock_logger):
        self.test_params["evalue"] = 0.001
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config._check_evalue()
        mock_logger.assert_not_called()

    def test_expand_all_paths_tilde(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.set("busco_run", "download_path", "~/test_download_path")
        config._expand_all_paths()
        self.assertEqual(
            config.get("busco_run", "download_path"),
            os.path.expanduser("~/test_download_path"),
        )

    def test_expand_all_paths_relative_path_current_dir(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.set("busco_run", "out_path", "./test_out_path")
        config._expand_all_paths()
        self.assertEqual(
            config.get("busco_run", "out_path"), os.path.abspath("./test_out_path")
        )

    def test_expand_all_paths_relative_path_parent_dir(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.set("busco_run", "in", "../test_input_file")
        config._expand_all_paths()
        self.assertEqual(
            config.get("busco_run", "in"), os.path.abspath("../test_input_file")
        )

    def test_expand_all_paths_hmmsearch(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.set("hmmsearch", "path", "~/test_hmmsearch_path")
        config._expand_all_paths()
        self.assertEqual(
            config.get("hmmsearch", "path"), os.path.expanduser("~/test_hmmsearch_path")
        )

    def test_required_input_exists_true(self):
        input_filename = "test_input_file"
        Path(input_filename).touch()
        self.test_params["in"] = input_filename
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        with self.assertLogs(BuscoConfig.logger, level="INFO"):
            config._check_required_input_exists()
        os.remove(input_filename)

    def test_required_input_exists_false(self):
        input_filename = "test_input_file"
        if os.path.exists(input_filename):
            os.remove(input_filename)
        self.test_params["in"] = input_filename
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        with self.assertRaises(SystemExit):
            config._check_required_input_exists()

    @patch("__main__.BuscoConfig_unittests.BuscoConfig.BuscoDownloadManager")
    def test_downloader_initialized(self, mock_downloader):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config._init_downloader()
        mock_downloader.assert_called()

    @patch("__main__.BuscoConfig_unittests.BuscoConfig.PrettyLog")
    def test_log_config(self, mock_pretty_log):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        with self.assertLogs(BuscoConfig.logger, level="DEBUG"):
            config.log_config()
        mock_pretty_log.assert_called()

    @patch.object(BuscoConfig.BuscoConfigMain, "log_config")
    @patch.object(BuscoConfig.BuscoConfigMain, "_init_downloader")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_required_input_exists")
    @patch.object(BuscoConfig.BuscoConfigMain, "_expand_all_paths")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_evalue")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_limit_value")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_out_value")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_allowed_keys")
    @patch.object(BuscoConfig.BuscoConfigMain, "_create_required_paths")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_no_previous_run")
    @patch.object(BuscoConfig.BuscoConfigMain, "_check_mandatory_keys_exist")
    def test_validation(
        self,
        mock_check_mandatory_keys,
        mock_check_no_previous_run,
        mock_create_required_paths,
        mock_check_allowed_keys,
        mock_check_out_value,
        mock_check_limit_value,
        mock_check_evalue,
        mock_expand_all_paths,
        mock_check_input,
        mock_init_downloader,
        mock_log_config,
    ):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.validate()
        mock_check_mandatory_keys.assert_called()
        mock_check_no_previous_run.assert_called()
        mock_create_required_paths.assert_called()
        mock_check_allowed_keys.assert_called()
        mock_check_out_value.assert_called()
        mock_check_limit_value.assert_called()
        mock_check_evalue.assert_called()
        mock_expand_all_paths.assert_called()
        mock_check_input.assert_called()
        mock_init_downloader.assert_called()
        mock_log_config.assert_called()

    def test_check_lineage_present_false(self):
        try:
            del self.test_params["lineage_dataset"]  # just in case, probably redundant
        except KeyError:
            pass
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        self.assertFalse(config.check_lineage_present())

    def test_check_lineage_present_true_with_dataset_version_correct(self):
        self.test_params["lineage_dataset"] = "test_dataset_odb10"
        self.test_params["datasets_version"] = "odb10"
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.check_lineage_present()
        self.assertEqual(
            config.get("busco_run", "datasets_version"),
            self.test_params["datasets_version"],
        )

    def test_check_lineage_present_true_with_dataset_version_mismatch(self):
        self.test_params["lineage_dataset"] = "test_dataset_odb10"
        self.test_params["datasets_version"] = "odb11"
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        with self.assertLogs(BuscoConfig.logger, level="WARNING"):
            config.check_lineage_present()
        self.assertEqual(
            config.get("busco_run", "datasets_version"),
            self.test_params["lineage_dataset"].split("_")[-1],
        )

    def test_check_lineage_present_true_with_odb_missing(self):
        self.test_params["lineage_dataset"] = "test_dataset"
        self.test_params["datasets_version"] = "odb10"
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        config.check_lineage_present()
        self.assertEqual(
            config.get("busco_run", "lineage_dataset"),
            "{}_{}".format(
                self.test_params["lineage_dataset"],
                self.test_params["datasets_version"],
            ),
        )

    def test_check_lineage_present_true_with_invalid_dataset_version(self):
        self.test_params["lineage_dataset"] = "test_dataset"
        self.test_params["datasets_version"] = "odb11"
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        with self.assertRaises(SystemExit):
            config.check_lineage_present()

    def test_set_results_dirname(self):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        test_dataset_path = "/path/to/lineage_dataset"
        with patch("busco.BuscoConfig.BuscoConfig.set"):
            config.set_results_dirname(test_dataset_path)
            config.set.assert_called_with(
                "busco_run", "lineage_results_dir", "run_lineage_dataset"
            )

    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfig")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    def test_autoconfig_init_propagates_mainconfig(self, mock_propagate, *args):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        BuscoConfig.BuscoConfigAuto(config, None)
        mock_propagate.assert_called_with(config)

    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfig")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    def test_autoconfig_init_sets_results_dirname(self, mock_set_dirname, *args):
        BuscoConfig.BuscoConfigAuto(None, "lineage")
        mock_set_dirname.assert_called_with("lineage")

    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfig")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    def test_autoconfig_init_creates_paths(self, mock_create_paths, *args):
        BuscoConfig.BuscoConfigAuto(None, None)
        mock_create_paths.assert_called()

    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfig")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    def test_autoconfig_init_downloads_lineage(self, mock_download_lineage, *args):
        BuscoConfig.BuscoConfigAuto(None, "lineage")
        mock_download_lineage.assert_called_with("lineage")

    @patch("busco.BuscoConfig.BuscoConfig")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    def test_autoconfig_init_loads_lineage_config(self, mock_load_dataset, *args):
        BuscoConfig.BuscoConfigAuto(None, None)
        mock_load_dataset.assert_called()

    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfig.__init__")
    def test_autoconfig_init_calls_super(self, mock_config_parent, *args):
        BuscoConfig.BuscoConfigAuto(None, None)
        mock_config_parent.assert_called()

    @patch("busco.BuscoConfig.BuscoConfigAuto._create_required_paths")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    def test_propagate_config(self, *args):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.params)
        config.configure()
        config.downloader = Mock()
        autoconfig = BuscoConfig.BuscoConfigAuto(config, "test")
        autoconfig._propagate_config(config)
        self.assertEqual(autoconfig, config)

    @patch("busco.BuscoConfig.BuscoConfigAuto.load_dataset_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.download_lineage_file")
    @patch("busco.BuscoConfig.BuscoConfigAuto._propagate_config")
    @patch("busco.BuscoConfig.BuscoConfigAuto.set_results_dirname")
    @patch("busco.BuscoConfig.BuscoConfigAuto.get", return_value="test")
    @patch("busco.BuscoConfig.BuscoConfig._create_required_paths")
    def test_autolineage_create_path_method_calls_parent(
        self, mock_create_paths, *args
    ):
        config = BuscoConfig.BuscoConfigMain(self.base_config, self.test_params)
        config.configure()
        BuscoConfig.BuscoConfigAuto(config, None)
        mock_create_paths.assert_called_with("test/auto_lineage")

    def tearDown(self):
        self.test_params = {}



import unittest
from unittest.mock import patch, call, Mock
from busco import AutoLineage, BuscoRunner


class TestAutoLineage(unittest.TestCase):
    def setUp(self):
        pass

    @patch("busco.BuscoConfig.BuscoConfigMain", autospec=True)
    def test_init_autolineage(self, mock_config_main):
        with self.assertLogs(AutoLineage.logger, 20):
            AutoLineage.AutoSelectLineage(mock_config_main)

    @patch("busco.AutoLineage.AutoSelectLineage.virus_check", return_value=False)
    @patch(
        "busco.BuscoConfig.BuscoConfigMain.getboolean",
        side_effect=[True, False, True, False, False],
    )
    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.os")
    @patch("busco.AutoLineage.BuscoRunner")
    @patch("busco.AutoLineage.AutoSelectLineage.get_best_match_lineage")
    @patch("busco.AutoLineage.AutoSelectLineage.run_lineages_list")
    @patch("busco.BuscoConfig.BuscoConfigMain", autospec=True)
    def test_run_auto_selector_lineage_lists_no_virus(
        self, mock_config_main, mock_run_lineages_list, *args
    ):
        for _ in range(3):
            asl = AutoLineage.AutoSelectLineage(mock_config_main)
            asl.selected_runner = Mock()
            asl.selected_runner.analysis.hmmer_runner.single_copy_buscos = [
                0
            ]  # avoid SystemExit with empty HMMER results
            asl.selected_runner.analysis.hmmer_runner.multi_copy_buscos = [0]
            asl.selected_runner.analysis.hmmer_runner.fragmented_buscos = [0]
            asl.run_auto_selector()

        calls = [
            call(["archaea", "bacteria"]),
            call(["eukaryota"]),
            call(["archaea", "bacteria", "eukaryota"]),
        ]
        mock_run_lineages_list.assert_has_calls(calls, any_order=True)

    @patch("busco.AutoLineage.logger.info")
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_run_lineages_initializes_BuscoConfigAuto(
        self, mock_config_main, mock_config_auto, *args
    ):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        test_lineages = ["a", "b", "c"]
        test_dataset_version = "<dataset_version>"
        asl.dataset_version = test_dataset_version
        asl.run_lineages_list(test_lineages)
        calls = [
            call(
                mock_config_main, "{}_{}".format(test_lineages[0], test_dataset_version)
            ),
            call(
                mock_config_main, "{}_{}".format(test_lineages[1], test_dataset_version)
            ),
            call(
                mock_config_main, "{}_{}".format(test_lineages[2], test_dataset_version)
            ),
        ]
        mock_config_auto.assert_has_calls(calls, any_order=True)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("busco.BuscoConfig.BuscoConfigMain")
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    def test_run_lineages_initializes_BuscoRunner(
        self, mock_runner, mock_config_main, mock_config_auto, *args
    ):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        test_lineages = ["a", "b", "c"]
        test_dataset_version = "<dataset_version>"
        asl.dataset_version = test_dataset_version
        asl.run_lineages_list(test_lineages)
        mock_runner.assert_called_with(mock_config_auto.return_value)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("busco.BuscoConfig.BuscoConfigMain")
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    def test_run_lineages_runs_analysis(self, mock_runner, mock_config_main, *args):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        test_lineages = ["a", "b", "c"]
        test_dataset_version = "<dataset_version>"
        asl.dataset_version = test_dataset_version
        asl.run_lineages_list(test_lineages)
        mock_runner.return_value.run_analysis.assert_called_with(
            callback=asl.record_results
        )

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("busco.BuscoConfig.BuscoConfigMain")
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    def test_run_lineages_returns_runners(self, mock_runner, mock_config_main, *args):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        test_lineages = ["a", "b", "c"]
        test_dataset_version = "<dataset_version>"
        asl.dataset_version = test_dataset_version
        runners = asl.run_lineages_list(test_lineages)
        self.assertEqual(
            runners,
            [
                mock_runner.return_value,
                mock_runner.return_value,
                mock_runner.return_value,
            ],
        )

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_record_results_first_run(self, mock_config_main, *args):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        asl.record_results(0, 1, 2, 0, 1, 2)
        self.assertGreater(len(asl.s_buscos), 0)
        self.assertGreater(len(asl.d_buscos), 0)
        self.assertGreater(len(asl.f_buscos), 0)
        self.assertGreater(len(asl.s_percents), 0)
        self.assertGreater(len(asl.d_percents), 0)
        self.assertGreater(len(asl.f_percents), 0)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.BuscoConfigAuto", autospec=True)
    @patch("__main__.AutoLineage_unittests.AutoLineage.BuscoRunner")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_record_results_multiple_runs(self, mock_config_main, *args):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        asl.record_results(0, 1, 2, 0, 1, 2)
        asl.record_results(0, 1, 2, 0, 1, 2)
        self.assertGreater(len(asl.s_buscos), 1)
        self.assertGreater(len(asl.d_buscos), 1)
        self.assertGreater(len(asl.f_buscos), 1)
        self.assertGreater(len(asl.s_percents), 1)
        self.assertGreater(len(asl.d_percents), 1)
        self.assertGreater(len(asl.f_percents), 1)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_evaluate_single_runner(self, mock_config_main, *args):
        runner1 = Mock()
        runner1.analysis.hmmer_runner.single_copy = 1
        runner1.analysis.hmmer_runner.multi_copy = 1
        runner1.analysis.hmmer_runner.only_fragments = 1
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        max_ind = asl.evaluate([runner1])
        self.assertEqual(max_ind, 0)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_evaluate_multiple_runners(self, mock_config_main, *args):
        runner1 = Mock()
        runner2 = Mock()
        runner3 = Mock()
        runner1.analysis.hmmer_runner.single_copy = 10
        runner1.analysis.hmmer_runner.multi_copy = 5
        runner2.analysis.hmmer_runner.single_copy = 15
        runner2.analysis.hmmer_runner.multi_copy = 5
        runner3.analysis.hmmer_runner.single_copy = 12
        runner3.analysis.hmmer_runner.multi_copy = 5
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        max_ind = asl.evaluate([runner1, runner2, runner3])
        self.assertEqual(max_ind, 1)

        runner2.analysis.hmmer_runner.single_copy = 10
        runner2.analysis.hmmer_runner.multi_copy = 6
        runner3.analysis.hmmer_runner.single_copy = 10
        runner3.analysis.hmmer_runner.multi_copy = 7
        max_ind = asl.evaluate([runner1, runner2, runner3])
        self.assertEqual(max_ind, 2)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_evaluate_first_order_tiebreak(self, mock_config_main, *args):
        runner1 = Mock()
        runner2 = Mock()
        runner3 = Mock()
        runner1.analysis.hmmer_runner.single_copy = 10
        runner1.analysis.hmmer_runner.multi_copy = 5
        runner1.analysis.hmmer_runner.only_fragments = 1
        runner2.analysis.hmmer_runner.single_copy = 10
        runner2.analysis.hmmer_runner.multi_copy = 5
        runner2.analysis.hmmer_runner.only_fragments = 2
        runner3.analysis.hmmer_runner.single_copy = 12
        runner3.analysis.hmmer_runner.multi_copy = 0
        runner3.analysis.hmmer_runner.only_fragments = 3
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        max_ind = asl.evaluate([runner1, runner2, runner3])
        self.assertEqual(max_ind, 1)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_evaluate_second_order_tiebreak(self, mock_config_main, *args):
        runner1 = Mock()
        runner2 = Mock()
        runner3 = Mock()
        runner4 = Mock()
        runner1.analysis.hmmer_runner.single_copy = 10
        runner1.analysis.hmmer_runner.multi_copy = 5
        runner1.analysis.hmmer_runner.only_fragments = 1
        runner1.analysis.hmmer_runner.s_percent = 20
        runner2.analysis.hmmer_runner.single_copy = 10
        runner2.analysis.hmmer_runner.multi_copy = 5
        runner2.analysis.hmmer_runner.only_fragments = 2
        runner2.analysis.hmmer_runner.s_percent = 40
        runner3.analysis.hmmer_runner.single_copy = 12
        runner3.analysis.hmmer_runner.multi_copy = 0
        runner3.analysis.hmmer_runner.only_fragments = 3
        runner3.analysis.hmmer_runner.s_percent = 60
        runner4.analysis.hmmer_runner.single_copy = 14
        runner4.analysis.hmmer_runner.multi_copy = 1
        runner4.analysis.hmmer_runner.only_fragments = 2
        runner4.analysis.hmmer_runner.s_percent = 80
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        max_ind = asl.evaluate([runner1, runner2, runner3, runner4])
        self.assertEqual(max_ind, 3)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain")
    def test_evaluate_third_order_tiebreak(self, mock_config_main, *args):
        runner1 = Mock()
        runner2 = Mock()
        runner3 = Mock()
        runner4 = Mock()
        runner1.analysis.hmmer_runner.single_copy = 10
        runner1.analysis.hmmer_runner.multi_copy = 5
        runner1.analysis.hmmer_runner.only_fragments = 1
        runner1.analysis.hmmer_runner.s_percent = 20
        runner2.analysis.hmmer_runner.single_copy = 10
        runner2.analysis.hmmer_runner.multi_copy = 5
        runner2.analysis.hmmer_runner.only_fragments = 2
        runner2.analysis.hmmer_runner.s_percent = 80
        runner3.analysis.hmmer_runner.single_copy = 12
        runner3.analysis.hmmer_runner.multi_copy = 0
        runner3.analysis.hmmer_runner.only_fragments = 3
        runner3.analysis.hmmer_runner.s_percent = 60
        runner4.analysis.hmmer_runner.single_copy = 14
        runner4.analysis.hmmer_runner.multi_copy = 1
        runner4.analysis.hmmer_runner.only_fragments = 2
        runner4.analysis.hmmer_runner.s_percent = 80
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        with self.assertLogs(AutoLineage.logger, "WARNING"):
            max_ind = asl.evaluate([runner1, runner2, runner3, runner4])
        self.assertEqual(max_ind, 1)

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.AutoLineage.AutoSelectLineage.cleanup_disused_runs")
    @patch("__main__.AutoLineage_unittests.BuscoRunner.BuscoRunner.mode_dict")
    @patch("busco.BuscoConfig.BuscoConfigMain", autospec=True)
    def test_get_best_match_lineage(
        self, mock_config_main, fake_modedict, mock_cleanup, *args
    ):
        mock_config_main.get.side_effect = [None]

        mock_config1 = Mock()
        mock_config2 = Mock()
        mock_config3 = Mock()
        mock_config1.get.side_effect = ["tran", None, "test1"]
        mock_config2.get.side_effect = ["tran", None, "test2"]
        mock_config3.get.side_effect = ["tran", None, "test3"]

        mock_analysis1 = Mock()
        mock_analysis2 = Mock()
        mock_analysis3 = Mock()
        mock_analysis1.return_value.hmmer_runner.single_copy = 75
        mock_analysis2.return_value.hmmer_runner.single_copy = 85
        mock_analysis3.return_value.hmmer_runner.single_copy = 80
        mock_analysis1.return_value.hmmer_runner.multi_copy = 5
        mock_analysis2.return_value.hmmer_runner.multi_copy = 6
        mock_analysis3.return_value.hmmer_runner.multi_copy = 7

        fake_modedict.__getitem__.side_effect = [
            mock_analysis1,
            mock_analysis2,
            mock_analysis3,
        ]

        runner1 = BuscoRunner.BuscoRunner(mock_config1)
        runner2 = BuscoRunner.BuscoRunner(mock_config2)
        runner3 = BuscoRunner.BuscoRunner(mock_config3)

        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        asl.get_best_match_lineage([runner1, runner2, runner3])
        self.assertEqual(asl.best_match_lineage_dataset, "test2")
        self.assertEqual(asl.selected_runner, runner2)
        mock_cleanup.assert_called_with([runner1, runner3])

    @patch("busco.AutoLineage.logger.info")
    @patch("busco.BuscoConfig.BuscoConfigMain", autospec=True)
    def test_cleanup_disused_runs(self, mock_config_main, *args):
        asl = AutoLineage.AutoSelectLineage(mock_config_main)
        mock_runner1 = Mock()
        mock_runner2 = Mock()
        mock_runner1.cleaned_up = False
        mock_runner2.cleaned_up = True
        asl.cleanup_disused_runs([mock_runner1, mock_runner2])
        mock_runner1.cleanup.assert_called()
        mock_runner2.cleanup.assert_not_called()

    # Todo: add tests for get_lineage_datasets, 3_dataset check and busco_placer step

    def tearDown(self):
        pass



#!/usr/bin/env python3
# coding: utf-8
"""

.. module:: generate_plot
   :synopsis: This module produces a graphic summary for BUSCO runs based on short summary files
.. versionadded:: 2.0.0
.. versionchanged:: 4.0.0

 This module produces a graphic summary for BUSCO runs based on short summary files

(``python3 generate_plot.py -h`` and user guide for details on how to do it)

Place the short summary files of all BUSCO runs you would like to see on the figure a single folder.
Keep the file named as follow: ``short_summary.[generic|specific].dataset.label.txt``, 'label' being used in the plot as species name

This tool produces the R code of the figure and uses ggplot2 (2.2.0+). If your system is able to run R, this script
automatically runs it.

You can find both the resulting R script for customisation and the figure in the working directory.

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""

import os
import sys
import time
import traceback
import argparse
import subprocess
import glob
from shutil import which
from argparse import RawTextHelpFormatter
import logging
from busco.BuscoLogger import BuscoLogger

#: working directory
_plot_dir = ""
#: r file name
_r_file = "busco_figure.R"

# to avoid running R
_no_r = False

#: Get an instance of _logger for keeping track of events
_logger = BuscoLogger.get_logger(__name__)

RCODE = (
    "######################################\n"
    "#\n"
    "# BUSCO summary figure\n"
    "# @version 4.0.0\n"
    "# @since BUSCO 2.0.0\n"
    "# \n"
    "# Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)\n"
    "# Licensed under the MIT license. See LICENSE.md file.\n"
    "#\n"
    "######################################\n"
    "\n"
    "# Load the required libraries\n"
    "library(ggplot2)\n"
    'library("grid")\n'
    "\n"
    "# !!! CONFIGURE YOUR PLOT HERE !!! \n"
    "# Output\n"
    'my_output <- paste(%s1,"busco_figure.png",sep="/") \n'
    "my_width <- 20\n"
    "my_height <- 15\n"
    'my_unit <- "cm"\n'
    "\n"
    "# Colors\n"
    'my_colors <- c("#56B4E9", "#3492C7", "#F0E442", "#F04442")\n'
    "# Bar height ratio\n"
    "my_bar_height <- 0.75\n"
    "\n"
    "# Legend\n"
    'my_title <- "BUSCO Assessment Results"\n'
    "\n"
    "# Font\n"
    'my_family <- "sans"\n'
    "my_size_ratio <- 1\n"
    "\n"
    "# !!! SEE YOUR DATA HERE !!! \n"
    "# Your data as generated by python, remove or add more\n"
    "my_species <- c%s2\n"
    "my_species <- factor(my_species)\n"
    "my_species <- factor(my_species,levels(my_species)[c(length(levels(my_species)):1)]) "
    "# reorder your species here just by changing the values in the vector :\n"
    "my_percentage <- c%s3\n"
    "my_values <- c%s4\n"
    "\n"
    "######################################\n"
    "######################################\n"
    "######################################\n"
    "# Code to produce the graph\n"
    "labsize = 1\n"
    "if (length(levels(my_species)) > 10){\n"
    " labsize = 0.66\n"
    "}\n"
    'print("Plotting the figure ...")\n'
    'category <- c(rep(c("S","D","F","M"),c%s5))\n'
    "category <-factor(category)\n"
    "category = factor(category,levels(category)[c(4,1,2,3)])\n"
    "df = data.frame(my_species,my_percentage,my_values,category)\n"
    "\n"
    "figure <- ggplot() + \n"
    "  \n"
    '  geom_bar(aes(y = my_percentage, x = my_species, fill = category), position = position_stack(reverse = TRUE), data = df, stat="identity", '
    "width=my_bar_height) + \n"
    "  coord_flip() + \n"
    "  theme_gray(base_size = 8) + \n"
    '  scale_y_continuous(labels = c("0","20","40","60","80","100"), breaks = c(0,20,40,60,80,100)) + \n'
    '  scale_fill_manual(values = my_colors,labels =c(" Complete (C) and single-copy (S)  ",\n'
    '                                                 " Complete (C) and duplicated (D)",\n'
    '                                                 " Fragmented (F)  ",\n'
    '                                                 " Missing (M)")) +   \n'
    "  ggtitle(my_title) + \n"
    '  xlab("") + \n'
    '  ylab("\\n%BUSCOs") + \n'
    "\n"
    '  theme(plot.title = element_text(family=my_family, hjust=0.5, colour = "black", size = rel(2.2)*my_size_ratio, face = '
    '"bold")) + \n'
    '  theme(legend.position="top",legend.title = element_blank()) + \n'
    "  theme(legend.text = element_text(family=my_family, size = rel(1.2)*my_size_ratio)) + \n"
    '  theme(panel.background = element_rect(color="#FFFFFF", fill="white")) + \n'
    "  theme(panel.grid.minor = element_blank()) + \n"
    "  theme(panel.grid.major = element_blank()) +\n"
    '  theme(axis.text.y = element_text(family=my_family, colour = "black", size = rel(1.66)*my_size_ratio)) + \n'
    '  theme(axis.text.x = element_text(family=my_family, colour = "black", size = rel(1.66)*my_size_ratio)) + \n'
    '  theme(axis.line = element_line(size=1*my_size_ratio, colour = "black")) + \n'
    '  theme(axis.ticks.length = unit(.85, "cm")) + \n'
    '  theme(axis.ticks.y = element_line(colour="white", size = 0)) + \n'
    '  theme(axis.ticks.x = element_line(colour="#222222")) + \n'
    '  theme(axis.ticks.length = unit(0.4, "cm")) + \n'
    "  theme(axis.title.x = element_text(family=my_family, size=rel(1.2)*my_size_ratio)) + \n"
    "  \n"
    "  guides(fill = guide_legend(override.aes = list(colour = NULL))) +\n"
    "  guides(fill=guide_legend(nrow=2,byrow=TRUE))\n"
    "  \n"
    "  for(i in rev(c(1:length(levels(my_species))))){\n"
    "    detailed_values <- my_values[my_species==my_species[my_species==levels(my_species)[i]]]\n"
    "    total_buscos <- sum(detailed_values)\n"
    "    figure <- figure + \n"
    '    annotate("text", label=paste("C:", detailed_values[1] + detailed_values[2], " [S:", detailed_values[1], '
    '", D:", detailed_values[2], "], F:", detailed_values[3], ", M:", detailed_values[4], ", n:", total_buscos, '
    'sep=""), \n'
    '             y=3, x = i, size = labsize*4*my_size_ratio, colour = "black", hjust=0, family=my_family)\n'
    "  }\n"
    "  \n"
    "ggsave(figure, file=my_output, width = my_width, height = my_height, unit = my_unit)\n"
    'print("Done")\n'
)


def _check_wd():
    """
    This function checks that the working directory exists with write permission
    :raises SystemExit: if the folder is absent or the user has no write permission
    """
    if not os.path.exists(_plot_dir):
        _logger.warning("Impossible to read %s" % _plot_dir)
        raise SystemExit()
    if not os.access(_plot_dir, os.W_OK):
        _logger.warning("Impossible to write into %s" % _plot_dir)
        raise SystemExit()


def _write_r_code(data):
    """
    This function write the R code in its own file
    :param data: the data loaded from the run folders used to generate the R file
    :type data: dict
    """
    r_file = open("%s%s" % (_plot_dir, _r_file), "w")
    r_file.write(
        RCODE.replace("%s1", '"%s"' % _plot_dir)
        .replace("%s2", str(tuple(data["species"])))
        .replace("%s3", str(tuple(data["percentages"])))
        .replace("%s4", str(tuple(data["values"])))
        .replace("%s5", "(1)")
    )


def _run_r_code():
    """
    This function runs the R code after it was generated
    It first checks that ggplot2 related libraries are present
    """
    # first try to load the two required package and warn the user if an error occur
    # package ggplot2
    need_to_exit = False
    ggplot2 = subprocess.Popen(
        ["R", "-e", "library(ggplot2)", "--quiet"],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    ggplot2_out = ggplot2.stderr.readlines() + ggplot2.stdout.readlines()
    if "Error" in str(ggplot2_out):
        _logger.warning(
            "Impossible to run R. The package ggplot2 does not seem to be installed. "
            "Please check your R installation. See also the --no_r option to avoid this message"
        )
        need_to_exit = True

    # package grid
    grid = subprocess.Popen(
        ["R", "-e", "library(grid)", "--quiet"],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    grid_out = grid.stderr.readlines() + grid.stdout.readlines()
    if "Error" in str(grid_out):
        _logger.warning(
            "Impossible to run R. The package grid does not seem to be installed. "
            "Please check your R installation. See also the --no_r option to avoid this message"
        )
        need_to_exit = True

    if need_to_exit:
        return None  # do not run the code, but no need to stop the execution

    # run R
    if which("Rscript") is not None:
        r_script = ["Rscript", "%s%s" % (_plot_dir, _r_file)]
        p = subprocess.Popen(
            r_script, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = p.communicate()
        if out:
            _logger.info("\n%s" % str(out.decode("utf-8")))
        if err:
            _logger.error("\n%s" % str(err.decode("utf-8")))
    else:
        _logger.error('"Rscript" is not accessible')
        raise SystemExit()


def _set_args():
    """
    This function sets the parameters provided by the user
    """
    parser = argparse.ArgumentParser(
        description="BUSCO plot generation tool.\n"
        "Place all BUSCO short summary files (short_summary.[generic|specific].dataset.label.txt) in a single folder. "
        "It will be "
        "your working directory, in which the generated plot files"
        " will be written"
        "\nSee also the user guide"
        " for additional information",
        usage="python3 generate_plot.py -wd [WORKING_DIRECTORY] [OTHER OPTIONS]",
        formatter_class=RawTextHelpFormatter,
        add_help=False,
    )

    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")

    required.add_argument(
        "-wd",
        "--working_directory",
        metavar="PATH",
        required=True,
        dest="working_directory",
        help="Define the location of your working directory",
    )
    optional.add_argument(
        "-rt",
        "--run_type",
        required=False,
        dest="run_type",
        help="type of summary to use, `generic` or `specific`",
    )
    optional.add_argument(
        "--no_r",
        help="To avoid to run R. It will just create the R script file in the working directory",
        action="store_true",
        dest="no_r",
    )
    optional.add_argument(
        "-q",
        "--quiet",
        help="Disable the info logs, displays only errors",
        action="store_true",
        dest="quiet",
    )
    optional.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit"
    )
    args = vars(parser.parse_args())
    if args["quiet"]:
        _logger.setLevel(logging.ERROR)
    if args["no_r"]:
        global _no_r
        _no_r = True
    global _plot_dir
    _plot_dir = args["working_directory"]
    if _plot_dir[-1] != "/":
        _plot_dir += "/"
    global _run_type
    _run_type = "*"
    if args["run_type"]:
        _run_type = args["run_type"]


def _load_data():
    """

    :return:
    """
    data = {"species": [], "values": [], "percentages": [], "species_tmp": []}
    datasets = set([])
    for f in glob.glob("%s/short_summary.%s.*.*.txt" % (_plot_dir, _run_type)):
        try:
            datasets.add(f.split("/")[-1].split(".")[1])
            content = open(f)
            comp = 0
            dupl = 0
            frag = 0
            miss = 0
            for line in content:
                if "Complete and single-copy BUSCOs" in line:
                    comp = int(line.split("\t")[1])
                elif "Complete and duplicated BUSCOs" in line:
                    dupl = int(line.split("\t")[1])
                elif "Fragmented BUSCOs" in line:
                    frag = int(line.split("\t")[1])
                elif "Missing BUSCOs" in line:
                    miss = int(line.split("\t")[1])
            data["species_tmp"] += [
                ".".join(
                    f.split("/")[-1].split(".")[3:-1] + [f.split("/")[-1].split(".")[2]]
                )
            ] * 4
            data["values"] += [comp, dupl, frag, miss]
            total = comp + dupl + frag + miss
            comp_pc = round(comp / float(total) * 100, 1)
            dupl_pc = round(dupl / float(total) * 100, 1)
            frag_pc = round(frag / float(total) * 100, 1)
            miss_pc = round(100 - comp_pc - dupl_pc - frag_pc, 1)
            data["percentages"] += [comp_pc, dupl_pc, frag_pc, miss_pc]
            _logger.info("Loaded %s successfully" % f)
        except IOError:
            _logger.warning("Impossible to use the file %s" % f)
    # if only one dataset, remove it from species label
    if len(datasets) == 1:
        data["species"] = [label.split(".")[0] for label in data["species_tmp"]]
    else:
        data["species"] = data["species_tmp"]
    if len(data["species"]) == 0:
        _logger.warning(
            "No files matching the pattern short_summary.%s were found in %s"
            % (_run_type, _plot_dir)
        )
        raise SystemExit()
    return data


def main():
    """
    This function produces a figure with all BUSCO runs present in the current folder
    """

    _set_args()  # Fetch the params provided by the user
    start_time = time.time()

    try:

        _logger.info(
            "****************** Start plot generation at %s ******************"
            % (time.strftime("%m/%d/%Y %H:%M:%S"))
        )

        # check working directory
        _check_wd()
        # load data
        _logger.info("Load data ...")
        data = _load_data()

        # write R code
        _logger.info("Generate the R code ...")
        _write_r_code(data)

        # run R code
        if not _no_r:
            _logger.info("Run the R code ...")
            _run_r_code()
        else:
            _logger.info("You chose not to run R")

        if not _logger.has_warning():
            _logger.info(
                "Plot generation done. Total running time: %s seconds"
                % str(time.time() - start_time)
            )
        else:
            _logger.info(
                "Plot generation done with WARNING(s). Total running time: %s seconds"
                % str(time.time() - start_time)
            )
        _logger.info("Results written in %s\n" % _plot_dir)

    except SystemExit:
        _logger.error("Plot generation failed !")
        _logger.info(
            "Check the logs, read the user guide, and check the BUSCO issue board on https://gitlab.com/ezlab/busco/issues"
        )
        raise

    except KeyboardInterrupt:
        _logger.error("A signal was sent to kill the process")
        _logger.error("Plot generation failed !")
        _logger.info(
            "Check the logs, read the user guide, and check the BUSCO issue board on https://gitlab.com/ezlab/busco/issues"
        )
        raise

    except BaseException:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        _logger.critical(
            "Unhandled exception occurred: %s"
            % repr(traceback.format_exception(exc_type, exc_value, exc_traceback))
        )
        _logger.error("Plot generation failed !\n")
        _logger.info(
            "Check the logs, read the user guide, and check the BUSCO issue board on https://gitlab.com/ezlab/busco/issues"
        )
        raise SystemExit()


# Entry point
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
"""
.. module:: BuscoDownloadManager
   :synopsis: BuscoDownloadManager manage the version and download the most recent file
.. versionadded:: 4.0.0
.. versionchanged:: 4.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""

import os
import time
import glob
import tarfile
import hashlib
import urllib.request
from urllib.error import URLError
import gzip

from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log

logger = BuscoLogger.get_logger(__name__)


class BuscoDownloadManager:
    """
    This class obtains and manages the version of data files required to run a BUSCO analysis.
    When the config parameter `offline` is set, no attempt to download is made.
    When the config parameter `auto_update_files` is set, new versions replace old versions.s
    Else, a warning is produced.
    """

    version_files = {}

    def __init__(self, config):
        """
        :param config: Values of all parameters to be used during the analysis
        :type config: BuscoConfig
        """
        self.offline = config.getboolean("busco_run", "offline")
        self.update_data = config.getboolean("busco_run", "update-data")
        self.download_base_url = config.get("busco_run", "download_base_url")
        self.local_download_path = config.get("busco_run", "download_path")
        self._create_main_download_dir()
        if not type(self).version_files and not self.offline:
            self._load_versions()

    def _create_main_download_dir(self):
        if not os.path.exists(self.local_download_path):
            # exist_ok=True to allow for multiple parallel BUSCO runs each trying to create this folder simultaneously
            os.makedirs(self.local_download_path, exist_ok=True)

    def _load_versions(self):
        try:
            versions_file = self._obtain_versions_file()
            with open(versions_file, "r") as v_file:
                for line in v_file:
                    line = line.strip().split("\t")
                    dataset_name = line[0]
                    dataset_date = line[1]
                    dataset_hash = line[2]
                    type(self).version_files.update(
                        {dataset_name: (dataset_date, dataset_hash)}
                    )
        except URLError as e:
            if self.offline:
                logger.warning(
                    "Unable to verify BUSCO datasets because of offline mode"
                )
            else:
                raise SystemExit(e)
        return

    @log("Downloading information on latest versions of BUSCO data...", logger)
    def _obtain_versions_file(self):
        remote_filepath = os.path.join(self.download_base_url, "file_versions.tsv")
        local_filepath = os.path.join(self.local_download_path, "file_versions.tsv")
        try:
            urllib.request.urlretrieve(remote_filepath, local_filepath)
        except URLError:
            SystemExit("Cannot reach {}".format(remote_filepath))
        return local_filepath

    def _create_category_dir(self, category):
        # if the category folder does not exist, create it
        category_folder = os.path.join(self.local_download_path, category)
        if not os.path.exists(category_folder):
            # exist_ok=True to allow for multiple parallel BUSCO runs, each trying to create this folder
            os.makedirs(category_folder, exist_ok=True)
        return

    @staticmethod
    def _extract_creation_date(dataset_config_file):
        dataset_date = None
        with open(dataset_config_file, "r") as data_config:
            for line in data_config:
                line = line.strip().split("=")
                if line[0] == "creation_date":
                    dataset_date = line[1]
                    break
        if not dataset_date:
            raise SystemExit(
                "Creation date could not be extracted from dataset.cfg file."
            )
        return dataset_date

    def _check_existing_version(self, local_filepath, category, data_basename):
        try:
            latest_update = type(self).version_files[data_basename][0]
        except KeyError:
            raise SystemExit(
                "{} is not a valid option for '{}'".format(data_basename, category)
            )
        path_basename, extension = os.path.splitext(data_basename)

        if category == "lineages":
            latest_version = ".".join([path_basename, latest_update])
            try:
                dataset_date = self._extract_creation_date(
                    os.path.join(local_filepath, "dataset.cfg")
                )
                up_to_date = dataset_date == latest_update
                present = True
            except FileNotFoundError:
                up_to_date = False
                present = False

        else:
            latest_version = ".".join(
                [path_basename, latest_update, extension.lstrip(".")]
            )
            local_filepath = local_filepath.replace(data_basename, latest_version)
            up_to_date = os.path.exists(local_filepath)
            path_to_check, extension_to_check = os.path.splitext(local_filepath)
            present = (
                len(
                    glob.glob(
                        "{}.*.{}".format(path_to_check[0:-11], extension_to_check[1:])
                    )
                )
                > 0
            )

        hash = type(self).version_files[data_basename][1]

        return present, up_to_date, latest_version, local_filepath, hash

    def get(self, data_name, category):
        if os.path.exists(data_name) and "/" in data_name:
            logger.info("Using local {} directory {}".format(category, data_name))
            return data_name
        elif "/" in data_name:
            raise SystemExit("{} does not exist".format(data_name))
        if self.offline:
            if category == "lineages":
                local_dataset = os.path.join(
                    self.local_download_path, category, data_name
                )
                if os.path.exists(local_dataset):
                    return local_dataset
                else:
                    raise SystemExit(
                        "Unable to run BUSCO in offline mode. Dataset {} does not "
                        "exist.".format(local_dataset)
                    )
            else:
                basename, extension = os.path.splitext(data_name)
                placement_files = sorted(
                    glob.glob(
                        os.path.join(
                            self.local_download_path,
                            category,
                            "{}.*{}".format(basename, extension),
                        )
                    )
                )
                if len(placement_files) > 0:
                    return placement_files[-1]
                    # todo: for offline mode, log which files are being used (in case of more than one glob match)
                else:
                    raise SystemExit(
                        "Unable to run BUSCO placer in offline mode. Cannot find necessary placement "
                        "files in {}".format(self.local_download_path)
                    )
        data_basename = os.path.basename(data_name)
        local_filepath = os.path.join(self.local_download_path, category, data_basename)
        (
            present,
            up_to_date,
            latest_version,
            local_filepath,
            hash,
        ) = self._check_existing_version(local_filepath, category, data_basename)

        if (not up_to_date and self.update_data) or not present:
            # download
            self._create_category_dir(category)
            compression_extension = ".tar.gz"
            remote_filepath = os.path.join(
                self.download_base_url, category, latest_version + compression_extension
            )
            if present and category == "lineages":
                self._rename_old_version(local_filepath)
            download_success = self._download_file(
                remote_filepath, local_filepath + compression_extension, hash
            )
            if download_success:
                local_filepath = self._decompress_file(
                    local_filepath + compression_extension
                )
                if present:
                    logger.warning(
                        "The file or folder {} was updated automatically.".format(
                            data_basename
                        )
                    )
        elif not up_to_date:
            logger.warning(
                "The file or folder {} is not the last available version. "
                "To update all data files to the last version, add the parameter "
                "--update-data in your next run.".format(local_filepath)
            )

        return local_filepath

    @staticmethod
    def _rename_old_version(local_filepath):
        if os.path.exists(local_filepath):
            try:
                os.rename(local_filepath, "{}.old".format(local_filepath))
                logger.info(
                    "Renaming {} into {}.old".format(local_filepath, local_filepath)
                )
            except OSError:
                try:
                    timestamp = time.time()
                    os.rename(
                        local_filepath, "{}.old.{}".format(local_filepath, timestamp)
                    )
                    logger.info(
                        "Renaming {} into {}.old.{}".format(
                            local_filepath, local_filepath, timestamp
                        )
                    )
                except OSError:
                    pass
        return

    @log("Downloading file {}", logger, func_arg=1)
    def _download_file(self, remote_filepath, local_filepath, expected_hash):
        try:
            urllib.request.urlretrieve(remote_filepath, local_filepath)
            observed_hash = type(self)._md5(local_filepath)
            if observed_hash != expected_hash:
                logger.error(
                    "md5 hash is incorrect: {} while {} expected".format(
                        str(observed_hash), str(expected_hash)
                    )
                )
                logger.info("deleting corrupted file {}".format(local_filepath))
                os.remove(local_filepath)
                raise SystemExit(
                    "BUSCO was unable to download or update all necessary files"
                )
            else:
                logger.debug("md5 hash is {}".format(observed_hash))
        except URLError:
            logger.error("Cannot reach {}".format(remote_filepath))
            return False
        return True

    @staticmethod
    def _md5(fname):
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @log("Decompressing file {}", logger, func_arg=1)
    def _decompress_file(self, local_filepath):
        unzipped_filename = local_filepath.replace(".gz", "")

        if os.path.splitext(local_filepath)[1] == ".gz":
            with gzip.open(os.path.join(local_filepath), "rb") as compressed_file:
                with open(unzipped_filename, "wb") as decompressed_file:
                    for line in compressed_file:
                        decompressed_file.write(line)
            os.remove(local_filepath)
            local_filepath = unzipped_filename

        if os.path.splitext(local_filepath)[1] == ".tar":
            untarred_filename = local_filepath.replace(".tar", "")
            with tarfile.open(local_filepath) as tar_file:
                tar_file.extractall(os.path.dirname(local_filepath))
            os.remove(local_filepath)
            local_filepath = untarred_filename

        return local_filepath
